Package: frheritage
Title: R Interface to Get French Heritage Data
Version: 0.1.1
Authors@R: 
    person("Matthieu", "Chevereau", , "matthieu.chevereau@hotmail.fr", role = c("aut", "cre"))
Description: Get spatial vector data from the Atlas du Patrimoine 
    (<http://atlas.patrimoines.culture.fr/atlas/trunk/>), 
    the official national platform of the French Ministry of Culture, 
    and facilitate its use within R geospatial workflows. 
    The package provides functions to list available heritage datasets, 
    query and retrieve heritage data using spatial queries based 
    on user-provided sf objects, perform spatial filtering operations, 
    and return results as sf objects suitable for spatial analysis, mapping, 
    and integration into heritage management and landscape studies.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: jsonlite, happign, httr2, sf, xml2
Depends: R (>= 4.1.0)
LazyData: true
Suggests: knitr, rmarkdown, testthat (>= 3.1.7), httptest2
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-17 16:46:21 UTC; matth
Author: Matthieu Chevereau [aut, cre]
Maintainer: Matthieu Chevereau <matthieu.chevereau@hotmail.fr>
Repository: CRAN
Date/Publication: 2026-04-17 17:00:02 UTC
Built: R 4.5.2; ; 2026-04-20 23:37:08 UTC; unix

Package: frheritage
Title: R Interface to Get French Heritage Data
Version: 0.1.1
Authors@R: 
    person("Matthieu", "Chevereau", , "matthieu.chevereau@hotmail.fr", role = c("aut", "cre"))
Description: Get spatial vector data from the Atlas du Patrimoine 
    (<http://atlas.patrimoines.culture.fr/atlas/trunk/>), 
    the official national platform of the French Ministry of Culture, 
    and facilitate its use within R geospatial workflows. 
    The package provides functions to list available heritage datasets, 
    query and retrieve heritage data using spatial queries based 
    on user-provided sf objects, perform spatial filtering operations, 
    and return results as sf objects suitable for spatial analysis, mapping, 
    and integration into heritage management and landscape studies.
License: GPL (>= 3)
Encoding: UTF-8
RoxygenNote: 7.3.3
Imports: jsonlite, happign, httr2, sf, xml2
Depends: R (>= 4.1.0)
LazyData: true
Suggests: knitr, rmarkdown, testthat (>= 3.1.7), httptest2
Config/testthat/edition: 3
NeedsCompilation: no
Packaged: 2026-04-17 16:46:21 UTC; matth
Author: Matthieu Chevereau [aut, cre]
Maintainer: Matthieu Chevereau <matthieu.chevereau@hotmail.fr>
Repository: CRAN
Date/Publication: 2026-04-17 17:00:02 UTC
Built: R 4.6.0; ; 2026-04-26 13:08:46 UTC; unix

# Generated by roxygen2: do not edit by hand

export(get_deps)
export(get_heritage)
export(get_heritage_ids)
export(get_heritage_layernames)
importFrom(happign,get_wfs)
importFrom(happign,intersects)
importFrom(httr2,req_method)
importFrom(httr2,req_perform)
importFrom(httr2,req_retry)
importFrom(httr2,req_timeout)
importFrom(httr2,request)
importFrom(httr2,resp_body_raw)
importFrom(httr2,resp_body_string)
importFrom(httr2,resp_status)
importFrom(jsonlite,toJSON)
importFrom(sf,st_bbox)
importFrom(sf,st_buffer)
importFrom(sf,st_cast)
importFrom(sf,st_centroid)
importFrom(sf,st_crs)
importFrom(sf,st_geometry)
importFrom(sf,st_geometry_type)
importFrom(sf,st_intersects)
importFrom(sf,st_is_empty)
importFrom(sf,st_is_valid)
importFrom(sf,st_join)
importFrom(sf,st_make_valid)
importFrom(sf,st_read)
importFrom(sf,st_set_crs)
importFrom(sf,st_sf)
importFrom(sf,st_transform)
importFrom(sf,st_union)
importFrom(sf,st_zm)
importFrom(utils,URLencode)
importFrom(utils,capture.output)
importFrom(utils,data)
importFrom(utils,unzip)
importFrom(xml2,read_xml)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_text)

# frheritage<img src="man/figures/frheritage_logo.png" align="right" height="138"/>

<!-- badges: start -->
[![CRAN status](https://www.r-pkg.org/badges/version/frheritage)](https://CRAN.R-project.org/package=frheritage) [![Lifecycle: experimental](https://img.shields.io/badge/lifecycle-experimental-orange.svg)](https://lifecycle.r-lib.org/articles/stages.html#experimental)
[![R-CMD-check](https://github.com/mucau/frheritage/actions/workflows/R-CMD-check.yaml/badge.svg)](https://github.com/mucau/frheritage/actions/workflows/R-CMD-check.yaml)
<!-- badges: end -->

## Overview

The **Atlas du Patrimoine** is the official national platform of the French Ministry of Culture that centralizes cultural heritage information. It provides access to a wide range of spatial and descriptive datasets, documenting France’s **architectural, archaeological, movable, and intangible heritage**.

The Atlas gathers and harmonizes data from several major databases such as:

-   Merimee – architectural and monumental heritage (churches, castles, etc.)

-   Palissy – movable heritage (objects, furniture, sculptures, etc.)

-   Patrimoine Mondial – UNESCO World Heritage sites in France

-   Patrimoine Archeologique – archaeological sites and surveys

-   Musees de France, Joconde, and other cultural inventories

All datasets are available through web services and open data exports provided by the Ministry of Culture. More details and an overview of available datasets can be found at: <http://atlas.patrimoines.culture.fr/atlas/trunk/static/presentation.html>

## Citation and source

Data are provided by the French Ministry of Culture, through the Atlas du Patrimoine platform: <http://atlas.patrimoines.culture.fr> Please cite both the Atlas du Patrimoine and this package when using data in scientific or professional work.

## API and data access disclaimer

The **frheritage** package is an independent, open-source client designed to facilitate access to public cultural data provided by the French Ministry of Culture. This package is **not affiliated with, endorsed by, or officially supported** by the Ministry.

All data remain the property of the **French Ministry of Culture** and are distributed under the **Etalab Open License 2.0** (or any equivalent open-data license indicated by the data provider). Users are responsible for complying with the applicable terms of use, attribution requirements, and update policies of the Atlas du Patrimoine datasets.

For official documentation and metadata, please refer to: <https://www.culture.gouv.fr/aides-demarches/protections-labels-et-appellations>

------------------------------------------------------------------------

## The frheritage package

### Purpose

The **frheritage** R package provides functions to explore, query, and retrieve datasets from the **Atlas du Patrimoine** within R. It aims to facilitate spatial analysis and data integration in geographic workflows, particularly for heritage management, cultural mapping, and landscape studies.

It allows to:

-   Retrieve and filter heritage datasets by code (internal nomenclature) and departement

-   Perform spatial queries directly from geometries (sf objects)

-   List all available heritage layers

-   Automatically validate user inputs and ensure compatibility with functions

### Main features

The main functions are:

-   `get_heritage_layernames()` : Returns a data frame listing all available heritage datasets with their internal codes, and labels. Useful for identifying which datasets are accessible through the package.

-    `get_heritage()`: Downloads one or several datasets from the **Atlas du Patrimoine** from a given `sf` object. It can apply spatial filters (intersects, within, etc.) and automatically handle geometries.

-    `get_heritage_ids()`: Retrieves available heritage IDs from the **Atlas du Patrimoine** from a given `sf` object.

The `get_heritage()` function works with the internal database `frheritage::all_ids` which contains (normally) all the IDs for metropolitan France.

### Getting start

``` r
# Just run
devtools::install_github("mucau/frheritage")

# Focus on a city
my_sf_polygon <- frcadastre::get_etalab(72191, data="communes")

# Retrieve the "immeuble classes ou inscrits" by using code "IMMH"
immh <- get_heritage(
  x = my_sf_polygon,
  data_code = "IMMH",
  buffer = 2500,
  spatial_filter = "within"
)

# To ask something else, try an other code !
# You don't know the codes ?
codes <- get_heritage_layernames()

# You don't trust my codes ?
# Okay, that's fine. You can check that I haven't forgotten any.
all_ids <- frheritage::all_ids # <- the internal database
ids <- get_heritage_ids(my_sf_polygon)

```

### Problems and Issues

-   Please report any issues or bugs you may encounter on the [dedicated page on github](https://github.com/mucau/frheritage/issues).

------------------------------------------------------------------------

Have fun ! :)

#' Get heritage layer codes and labels
#'
#' This function provides a reference table mapping internal codes to
#' explore heritage layer available datasets from French Ministry
#' of Culture's "Atlas du Patrimoine".
#' Optionally, it can filter the table to return only selected codes.
#'
#' @param code Optional `character` vector of layer codes to filter.
#' If `NULL`, the full reference table is returned.
#'
#' @details
#' The available codes and their meanings are:
#' \itemize{
#'   \item `IMDN` — Domaines nationaux
#'   \item `IMMH` — Immeubles classes ou inscrits
#'   \item `IMUN` — Patrimoine Mondial UNESCO – Emprise surfacique des biens
#'   \item `LACR` — Architecture Contemporaine Remarquable
#'   \item `PADN` — Protection au titre des abords de domaines nationaux
#'   \item `PAMH` — Protection au titre des abords de monuments historiques
#'   \item `PAUN` — Patrimoine Mondial UNESCO – Zone tampon des biens
#'   \item `SICI` — Sites classes ou inscrits
#'   \item `SIPR` — Sites patrimoniaux remarquables
#'   \item `ZPPA` — Zones de presomption de prescriptions archeologiques
#' }
#'
#' @return A `data.frame` with two columns:
#' \describe{
#'   \item{code}{Official heritage layer code (character).}
#'   \item{label}{Descriptive label (character).}
#' }
#'
#' @examples
#' # Get the full table of heritage layer codes
#' get_heritage_layernames()
#'
#' # Filter for specific codes
#' get_heritage_layernames(c("IMMH", "SICI"))
#'
#' @export
#' @importFrom utils data
get_heritage_layernames <- function(code = NULL) {
  # Reference table of heritage layer codes and labels
  datas <- data.frame(
    code = c("IMDN", "IMMH", "IMUN", "LACR", "PADN", "PAMH", "PAUN", "SICI", "SIPR", "ZPPA"),
    label = c(
      "Domaines nationaux",
      "Immeubles classes ou inscrits",
      "Patrimoine Mondial UNESCO - Emprise surfacique des biens",
      "Architecture Contemporaine Remarquable",
      "Protection au titre des abords de domaines nationaux",
      "Protection au titre des abords de monuments historiques",
      "Patrimoine Mondial UNESCO - Zone tampon des biens",
      "Sites classes ou inscrits",
      "Sites patrimoniaux remarquables",
      "Zones de presomption de prescriptions archeologiques"
    ),
    stringsAsFactors = FALSE
  )

  # Return the full table if no filtering is requested
  if (is.null(code)) {
    return(datas)
  }

  # Validate provided codes
  invalid <- setdiff(code, datas$code)
  if (length(invalid) > 0) {
    stop(sprintf("Invalid code(s): %s", paste(invalid, collapse = ", ")), call. = FALSE)
  }

  # Return only the requested rows
  datas[datas$code %in% code, , drop = FALSE]
}

test_that("get_heritage_layernames() returns the full table when no code is provided", {
  df <- get_heritage_layernames()
  expect_s3_class(df, "data.frame")
  expect_true(all(c("code", "label") %in% colnames(df)))
  expect_true(nrow(df) >= 1)
})

test_that("get_heritage_layernames() returns only requested valid codes", {
  df <- get_heritage_layernames(c("IMDN", "IMMH"))
  expect_s3_class(df, "data.frame")
  expect_equal(df$code, c("IMDN", "IMMH"))
})

test_that("get_heritage_layernames() errors on invalid code", {
  expect_error(get_heritage_layernames("INVALID"), regexp = "Invalid code")
})

test_that("get_heritage_layernames() errors when part of codes are invalid", {
  expect_error(get_heritage_layernames(c("IMDN", "INVALID")), regexp = "Invalid code")
})

#' Available layer datasets
#'
#' A dataset containing available layer datasets from the French
#' Ministry of Culture's "Atlas du Patrimoine" service.
#'
#' @format A `data.frame` with 5 columns:
#' \describe{
#'   \item{id}{(character) Layer id.}
#'   \item{title}{(character) Layer title.}
#'   \item{guid}{(integer) Layer guid.}
#'   \item{code}{(character) Internal code, determined from tile.}
#'   \item{departement}{(character) Layer department}
#' }
#'
#' @details
#' Contains only data available on the 96 departments of metropolitan France.
#'
"all_ids"

#' Example files for testing
#'
#' These files are used internally for testing functions.
#'
#' @name extdata_examples
#' @docType data
#' @keywords internal
NULL

#' Execute an expression completely silently
#'
#' This utility executes any R expression or function call and returns its value,
#' while suppressing all printed output (cat, print), warnings, and messages.
#'
#' @param expr An R expression or function call to execute silently.
#'
#' @return The result of evaluating `expr`.
#'
#' @details
#' - Useful to suppress noisy outputs from functions during package or script execution.
#' - Evaluation occurs in the parent environment to access local variables.
#' - Works cross-platform by redirecting `capture.output` to `nul` on Windows or `/dev/null` on other OS.
#'
#' @importFrom utils capture.output
#'
#' @keywords internal
#'
quiet <- function(expr) {
  # Determine null file for capture.output based on OS
  null_file <- if (.Platform$OS.type == "windows") "nul" else "/dev/null"

  # Evaluate in the parent environment to access local variables
  result <- capture.output(
    value <- suppressWarnings(suppressMessages(eval(substitute(expr), envir = parent.frame()))),
    file = null_file
  )

  # Return the actual result
  value
}

#' Check if the Atlas service is reachable
#'
#' Performs a lightweight HTTP request to test whether a remote service
#' (typically the French cultural heritage Atlas) is reachable within a
#' specified timeout. The function uses a `HEAD` request to avoid downloading
#' content and returns a logical indicating availability.
#'
#' This function is designed to be robust and silent: any network error
#' (timeout, DNS failure, connection refusal, etc.) results in `FALSE`.
#'
#' @param url A `character` string of length 1. The URL of the service to test.
#'   Default is `"http://atlas.patrimoines.culture.fr/atlas/trunk/"`.
#' @param timeout Numeric. Maximum time in seconds allowed for the request
#'   (including connection and response). Default is `4`.
#'
#' @return A `logical` scalar:
#' \itemize{
#'   \item `TRUE` if the service responds with an HTTP status code < 400
#'   \item `FALSE` otherwise (error, timeout, or HTTP status >= 400)
#' }
#'
#' @details
#' The function relies on \pkg{httr2} and internally performs a `HEAD` request
#' using `httr2::req_perform()`. It does not modify global options and
#' does not raise errors.
#'
#' Note that this function checks **service reachability**, not full
#' functional correctness of the endpoint.
#'
#' @importFrom httr2 request req_method req_timeout req_perform resp_status
#'
#' @keywords internal
#'
atlas_ok <- function(url = "http://atlas.patrimoines.culture.fr/atlas/trunk/",
                     timeout = 4) {
  tryCatch({

    resp <- httr2::request(url) |>
      httr2::req_method("HEAD") |>
      httr2::req_timeout(timeout) |>
      httr2::req_perform()

    httr2::resp_status(resp) < 400

  }, error = function(e) FALSE)
}

test_that("geo_extent() computes bounding boxes correctly", {
  skip_if_not_installed("sf")

  # Load test data from package
  sevres_path <- system.file("extdata", "sevres.rda", package = "frheritage")
  hauts_path <- system.file("extdata", "hauts_de_seine.rda", package = "frheritage")

  skip_if(!file.exists(sevres_path), "sevres.rda not found")
  skip_if(!file.exists(hauts_path), "hauts_de_seine.rda not found")

  load(sevres_path)
  load(hauts_path)

  # Ensure objects are sf
  expect_s3_class(sevres, "sf")
  expect_s3_class(hauts_de_seine, "sf")

  # Run function silently
  bb_sevres <- quiet(geo_extent(sevres))
  bb_hauts  <- quiet(geo_extent(hauts_de_seine, crs = 2154))

  # Check output structure
  expect_s3_class(bb_sevres, "data.frame")
  expect_s3_class(bb_hauts, "data.frame")

  expect_named(bb_sevres, c("left", "bottom", "right", "top"))
  expect_named(bb_hauts, c("left", "bottom", "right", "top"))

  # Check number of rows matches input features
  expect_equal(nrow(bb_sevres), nrow(sevres))
  expect_equal(nrow(bb_hauts), nrow(hauts_de_seine))

  # Check bounding box validity (min <= max)
  expect_true(all(bb_sevres$left <= bb_sevres$right))
  expect_true(all(bb_sevres$bottom <= bb_sevres$top))
  expect_true(all(bb_hauts$left <= bb_hauts$right))
  expect_true(all(bb_hauts$bottom <= bb_hauts$top))

  # Check CRS transformation consistency (global reference bbox)
  sevres_4326 <- sf::st_transform(sevres, 4326)
  bb_ref <- sf::st_bbox(sevres_4326)

  expect_true(all(bb_sevres$left >= bb_ref["xmin"]))
  expect_true(all(bb_sevres$right <= bb_ref["xmax"]))
  expect_true(all(bb_sevres$bottom >= bb_ref["ymin"]))
  expect_true(all(bb_sevres$top <= bb_ref["ymax"]))
})

test_that("quiet() returns correct value", {
  expect_identical(quiet(1 + 1), 2)
})

test_that("quiet() suppresses messages and warnings", {
  f <- function() { message("msg"); warning("warn"); 42 }
  output <- capture.output({
    res <- quiet(f())
  })
  expect_identical(res, 42)
  expect_length(output, 0)
})

test_that("quiet() evaluates in parent environment", {
  x <- 10
  expect_identical(quiet(x * 2), 20)
})

test_that("quiet() handles multi-line expressions", {
  res <- quiet({
    a <- 2; b <- 5
    a * b
  })
  expect_identical(res, 10)
})

test_that("quiet() works with sf operations", {
  skip_if_not_installed("sf")

  # Load test data
  sevres_path <- system.file("extdata", "sevres.rda", package = "frheritage")
  hauts_de_seine_path <- system.file("extdata", "hauts_de_seine.rda", package = "frheritage")

  skip_if(!file.exists(sevres_path), "sevres.rda not found in inst/extdata")
  skip_if(!file.exists(hauts_de_seine_path), "hauts_de_seine.rda not found in inst/extdata")

  load(sevres_path)
  load(hauts_de_seine_path)

  # Ensure sf objects are available and valid
  expect_true(exists("sevres"))
  expect_true(exists("hauts_de_seine"))
  expect_s3_class(sevres, "sf")
  expect_s3_class(hauts_de_seine, "sf")

  # Test 1: st_cast() inside quiet
  output_cast <- capture.output({
    res_cast <- quiet(sf::st_cast(sevres, "POLYGON"))
  })
  expect_s3_class(res_cast, "sf")
  expect_length(output_cast, 0)  # should produce no output

  # Test 2: st_intersection() inside quiet
  output_inter <- capture.output({
    res_inter <- quiet(sf::st_intersection(sf::st_transform(sevres, 2154),
                                                sf::st_transform(hauts_de_seine, 2154)))
  })
  expect_s3_class(res_inter, "sf")
  expect_length(output_inter, 0)
})

test_that("get_deps() works", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.209271, 48.82095)), crs = 4326))
  res <- sf::st_sf(
    code_insee = "92",
    geometry = sf::st_sfc(
      sf::st_polygon(list(rbind(
        c(2.18, 48.80),
        c(2.24, 48.80),
        c(2.24, 48.85),
        c(2.18, 48.85),
        c(2.18, 48.80)
      ))),
      crs = 4326
    )
  )

  testthat::local_mocked_bindings(
    get_wfs = function(...) res,
    .package = "happign"
  )

  dep <- get_deps(x)

  # Checks
  expect_true(is.character(dep))
  expect_true("92" %in% dep)
  expect_length(dep, 1)
})

test_that("get_deps() returns null when no data", {
  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(5, 5)), crs = 2154))
  res <- sf::st_sf(code_insee = character(0),
                   geometry = sf::st_sfc(crs = 4326))

  testthat::local_mocked_bindings(
    get_wfs = function(...) res,
    .package = "happign"
  )

  dep <- quiet(get_deps(x))

  # tests
  expect_null(dep)
})

test_that("get_deps() handles WFS error", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2, 48)), crs = 4326))

  testthat::local_mocked_bindings(
    get_wfs = function(...) stop("WFS error"),
    .package = "happign"
  )

  expect_message(
    res <- get_deps(x),
    "get_wfs\\(\\) failed"
  ) |> quiet()

  expect_null(res)
})

# mock tests ----
test_that("atlas_ok returns TRUE on HTTP 200", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      httr2::response(
        status_code = 200,
        headers = list(),
        body = raw()
      )
    },
    .package = "httr2"
  )

  expect_true(atlas_ok("http://fake.url"))
})

test_that("atlas_ok returns TRUE on HTTP 302", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      httr2::response(
        status_code = 302,
        headers = list(),
        body = raw()
      )
    },
    .package = "httr2"
  )

  expect_true(atlas_ok("http://fake.url"))
})

test_that("atlas_ok returns FALSE on HTTP 500", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      httr2::response(
        status_code = 500,
        headers = list(),
        body = raw()
      )
    },
    .package = "httr2"
  )

  expect_false(atlas_ok("http://fake.url"))
})

test_that("atlas_ok returns FALSE on timeout", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      stop(structure(list(), class = c("httr2_timeout", "error")))
    },
    .package = "httr2"
  )

  expect_false(atlas_ok("http://fake.url"))
})

test_that("atlas_ok returns FALSE on generic error", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      stop("connection failed")
    },
    .package = "httr2"
  )

  expect_false(atlas_ok("http://fake.url"))
})

test_that("atlas_ok returns a single logical", {

  testthat::local_mocked_bindings(
    req_perform = function(req) {
      httr2::response(200, list(), raw())
    },
    .package = "httr2"
  )

  res <- atlas_ok("http://fake.url")

  expect_type(res, "logical")
  expect_length(res, 1)
})

# real tests ----
test_that("atlas_ok real network behavior", {

  skip_on_cran()
  skip_on_ci()
  skip_if_offline()

  # OK
  expect_true(
    atlas_ok("https://www.google.com", timeout = 3)
  )

  # DNS invalide
  expect_false(
    atlas_ok("http://example.invalid", timeout = 2)
  )

  # Timeout
  expect_false(
    atlas_ok("http://10.255.255.1", timeout = 2)
  )

  # HTTP error
  expect_false(
    atlas_ok("https://httpbin.org/status/500", timeout = 3)
  )
})

test_that("geo_prepare() prepares and aggregates sf objects correctly", {
  skip_if_not_installed("sf")

  # Load test data from package
  sevres_path <- system.file("extdata", "sevres.rda", package = "frheritage")
  hauts_path <- system.file("extdata", "hauts_de_seine.rda", package = "frheritage")
  skip_if(!file.exists(sevres_path), "sevres.rda not found")
  skip_if(!file.exists(hauts_path), "hauts_de_seine.rda not found")
  load(sevres_path)
  load(hauts_path)

  # Ensure objects are sf
  expect_s3_class(sevres, "sf")
  expect_s3_class(hauts_de_seine, "sf")

  # Run function silently on both datasets
  output <- capture.output({
    res_sevres <- quiet(geo_prepare(sevres, crs = 2154, buffer = 50))
    res_hauts  <- quiet(geo_prepare(hauts_de_seine, crs = 2154, buffer = 50))
  })
  expect_length(output, 0)

  # Returned objects are sf
  expect_s3_class(res_sevres, "sf")
  expect_s3_class(res_hauts, "sf")

  # Geometry type is POLYGON after aggregation
  expect_true(all(sf::st_geometry_type(res_sevres) == "POLYGON"))
  expect_true(all(sf::st_geometry_type(res_hauts) == "POLYGON"))

  # Geometries are non-empty
  expect_false(any(sf::st_is_empty(res_sevres)))
  expect_false(any(sf::st_is_empty(res_hauts)))

  # CRS is correctly set
  expect_equal(sf::st_crs(res_sevres)$epsg, 2154)
  expect_equal(sf::st_crs(res_hauts)$epsg, 2154)

  # Aggregated geometries should reduce the number of features
  expect_true(nrow(res_sevres) <= nrow(sevres))
  expect_true(nrow(res_hauts) <= nrow(hauts_de_seine))
})

test_that("get_heritage() returns sf object with valid mocked data", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,

    get_deps = function(x) c("92"),

    zip_query_build = function(...) "http://fake-url.zip",
    zip_download = function(...) system.file("extdata/file4d441aa75cf1.zip", package = "frheritage"),

    geo_shapefile_read = function(zip_tmp, crs) {
      sf::st_sf(
        id = "1",
        geometry = sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = crs)
      )
    },

    geo_sf_bind = function(lst) do.call(rbind, lst),

    .package = "frheritage"
  )

  res <- get_heritage(x, data_code = "IMMH", verbose = FALSE)

  expect_s3_class(res, "sf")
  expect_equal(nrow(res), 1)
})

test_that("get_heritage() stops if extent is invalid", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    geo_extent = function(y) NULL,
    get_deps = function(x) c("92"),
    .package = "frheritage"
  )

  expect_error(
    get_heritage(x, data_code = "IMMH"),
    "Invalid extents"
  )
})

test_that("get_heritage() stops if no department found", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(x) NULL,
    .package = "frheritage"
  )

  expect_error(
    get_heritage(x, data_code = "IMMH"),
    "At least one geometry has no department"
  )
})

test_that("get_heritage() returns empty sf and warning if download fails", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(x) c("92"),
    zip_query_build = function(...) "http://fake-url.zip",
    zip_download = function(...) NULL,
    .package = "frheritage"
  )

  res <- get_heritage(x, data_code = "IMMH", verbose = FALSE)

  expect_s3_class(res, "sf")
  expect_equal(nrow(res), 0)
})

test_that("get_heritage() fails with multiple data_code", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    .package = "frheritage"
  )

  expect_error(
    get_heritage(x, data_code = c("A", "B")),
    "`data_code` must be one of:"
  )
})

test_that("get_heritage() returns empty sf if no IDs found", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) "92",
    data_filter = function(...) data.frame(),
    .package = "frheritage"
  )

  res <- get_heritage(x, data_code = "IMMH", verbose = FALSE)

  expect_s3_class(res, "sf")
  expect_equal(nrow(res), 0)
})

test_that("get_heritage() handles no matching code rows", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) "92",
    data_filter = function(...) {
      data.frame(
        id = "1",
        title = "t",
        guid = "g",
        departement = "OTHER"
      )
    },
    .package = "frheritage"
  )

  res <- get_heritage(x, data_code = "IMMH", verbose = FALSE)

  expect_s3_class(res, "sf")
  expect_equal(nrow(res), 0)
})

test_that("get_heritage() returns empty sf if all downloads fail", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) "92",
    zip_download = function(...) NULL,
    geo_sf_bind = function(...) NULL,
    .package = "frheritage"
  )

  res <- get_heritage(x, data_code = "IMMH", verbose = FALSE)

  expect_equal(nrow(res), 0)
})

test_that("get_heritage() warns on partial download failure", {

  x <- sf::st_as_sf(sf::st_sfc(sf::st_point(c(2.2, 48.8)), crs = 4326))

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) "92",
    zip_download = function(url, id) if (id == "1") "ok" else NULL,
    geo_shapefile_read = function(...) sf::st_as_sf(sf::st_sfc(sf::st_point(c(0,0)), crs = 2154)),
    geo_sf_bind = function(lst) do.call(rbind, lst),
    .package = "frheritage"
  )

  expect_message(
    res <- get_heritage(x, data_code = "IMMH", verbose = TRUE),
    "could not be retrieved|Download failed|Empty/invalid"
  ) |> quiet()

  expect_s3_class(res, "sf")
})

# Query to service service-
#' Build a search URL for the RSS service
#'
#' This internal helper constructs a URL to query for the French Ministry
#' of Culture's "Atlas du Patrimoine" RSS service.
#' The URL filters data using a bounding box (`extent`) and an
#' INSEE department code (`insee_dep`).
#'
#' @param extent A named list with elements `left`, `bottom`, `right`, `top`
#' defining the bounding box coordinates.
#' @param insee_dep A character or numeric INSEE department code.
#'
#' @return A character string containing the full URL to query the service.
#'
#' @importFrom utils URLencode
#'
#' @keywords internal
#'
ids_url_build <- function(extent, insee_dep) {
  # Validate input structure
  required <- c("left", "bottom", "right", "top")
  missing <- setdiff(required, names(extent))
  if (length(missing) > 0) {
    stop("Missing extent element(s): ", paste(missing, collapse = ", "), call. = FALSE)
  }

  # Base URL
  base_url <- "http://atlas.patrimoines.culture.fr/geosource/srv/fr/rss.search?"

  # Build query parameters
  query_params <- list(
    "_dc" = as.character(as.numeric(Sys.time()) * 1000),
    from = 1,
    to = 30,
    sortby = "MCC",
    pertinentScaleLevel = 1,
    georss = "simple",
    westBL = extent$left,
    eastBL = extent$right,
    southBL = extent$bottom,
    northBL = extent$top,
    geoForm = "BBOX",
    geoID = paste0("DEP_", insee_dep),
    themekeywords = "Protection"
  )

  # Encode query string
  query_string <- paste(
    names(query_params),
    vapply(query_params, function(x) URLencode(as.character(x), reserved = TRUE), character(1)),
    sep = "=",
    collapse = "&"
  )

  # Combine with proxy URL
  paste0(
    "http://atlas.patrimoines.culture.fr/atlas/trunk/proxy_ign.php?url=",
    URLencode(paste0(base_url, query_string), reserved = TRUE)
  )
}

#' Download and parse heritage metadata from a URL
#'
#' This internal helper performs an HTTP GET request to the given URL
#' (typically generated by [ids_url_build()]) and parses the RSS XML
#' response into a structured `data.frame`.
#'
#' @param url `Character.` The complete request URL.
#' @param timeout `Numeric.` Request timeout in seconds (default = `60`).
#' @return
#' A `data.frame` with columns:
#' \describe{
#'   \item{id}{Numeric identifier extracted from the GUID.}
#'   \item{title}{Record title as published in the service.}
#'   \item{guid}{Full GUID (unique resource identifier).}
#' }
#' Returns an empty `data.frame` if the request fails or no items are found.
#'
#' @importFrom httr2 request req_timeout req_perform resp_status resp_body_string req_retry
#' @importFrom xml2 read_xml xml_find_all xml_text
#'
#' @keywords internal
#'
ids_download <- function(url, timeout = 60) {
  # Build request
  req <- httr2::request(url) |>
    httr2::req_timeout(timeout) |>
    httr2::req_retry(
      max_tries = 2,
      backoff = ~5,
      is_transient = function(resp) {
        httr2::resp_status(resp) >= 500
      }
    ) |>
    quiet()

  # Perform with retry
  res <- tryCatch(
    httr2::req_perform(req),
    error = function(e) {
      warning("Request failed: ", e$message)
      data.frame()
    }
  )

  # Check response
  if (is.null(res) || httr2::resp_status(res) != 200) {
    warning("Invalid or empty response.")
    return(data.frame())
  }

  # Parse XML
  xml_txt <- httr2::resp_body_string(res)

  # Retrieve doc
  doc <- tryCatch(
    xml2::read_xml(xml_txt),
    error = function(e) {
      warning("XML parsing failed: ", e$message)
      return(NULL)
    }
  )
  if (is.null(doc)) return(data.frame())

  # Retrieve items
  items <- xml2::xml_find_all(doc, ".//item")
  if (length(items) == 0) {
    warning("No items found.")
    return(data.frame())
  }

  # Extract data
  rows <- lapply(items, function(item) {

    title <- xml2::xml_text(xml2::xml_find_first(item, "./title"))
    guid  <- xml2::xml_text(xml2::xml_find_first(item, "./guid"))

    id <- sub(".*MD_([0-9]+).*", "\\1", guid)

    data.frame(
      id = id,
      title = title,
      guid = guid,
      stringsAsFactors = FALSE
    )
  })

  df <- do.call(rbind, rows)

  if (nrow(df) == 0 || all(is.na(df$id))) {
    warning("No valid IDs extracted.")
    return(data.frame())
  }

  return(df)
}


#' Determine internal codes from IDs titles
#'
#' This internal function converts IDs tiles into their corresponding
#' internal codes (e.g., "IMMH", "PAMH").
#' It matches based on predefined keyword patterns and chooses the most
#' specific code when multiple patterns match.
#'
#' @param x `character` vector of IDs titles.
#'
#' @return `character` vector of internal codes corresponding to each input IDs titles.
#' Returns `NA` if no pattern matches.
#'
#' @keywords internal
#'
ids_to_codes <- function(x) {
  # Load patterns from internal data
  patterns <- heritage_patterns

  # Internal helper to find the best matching code for a given description
  find_code <- function(text) {
    matches <- character(0)
    matched_lengths <- integer(0)

    for (code in names(patterns)) {
      pats <- patterns[[code]]
      hit <- sapply(pats, grepl, text, ignore.case = TRUE)
      if (any(hit)) {
        # Keep track of matching code and length of the longest matching pattern
        longest <- max(nchar(pats[hit]))
        matches <- c(matches, code)
        matched_lengths <- c(matched_lengths, longest)
      }
    }

    if (length(matches) == 0) return(NA_character_)
    # Select the code with the longest matching pattern (more specific)
    matches[which.max(matched_lengths)]
  }

  # Apply to each element of the input vector
  sapply(x, find_code, USE.NAMES = FALSE)
}

test_that("ids_to_codes() returns correct codes for exact matches", {
  expect_identical(ids_to_codes("Immeubles classés ou inscrits"), "IMMH")
  expect_identical(ids_to_codes("Protection au titre des abords de monuments historiques"), "PAMH")
  expect_identical(ids_to_codes("Domaines nationaux"), "IMDN")
})

test_that("ids_to_codes() handles case-insensitive matches", {
  expect_identical(ids_to_codes("immeubles classés ou inscrits"), "IMMH")
  expect_identical(ids_to_codes("DOMAINES NATIONAUX"), "IMDN")
})

test_that("ids_to_codes() selects the most specific pattern when multiple match", {
  expect_identical(ids_to_codes("Abord d'un monument historique"), "PAMH")
})

test_that("ids_to_codes() returns NA when no pattern matches", {
  expect_true(is.na(ids_to_codes("Unknown heritage site")))
  expect_true(all(is.na(ids_to_codes(c("", "random text")))))
})

test_that("ids_to_codes() works for a vector of mixed titles", {
  titles <- c(
    "Immeubles classés ou inscrits",
    "Sites patrimoniaux remarquables",
    "Non-matching title",
    "Domaines nationaux"
  )
  codes <- ids_to_codes(titles)
  expect_identical(codes, c("IMMH", "SIPR", NA_character_, "IMDN"))
})

test_that("ids_to_codes() handles very long text containing a pattern", {
  long_text <- paste(rep("bla ", 50), "Sites inscrits", rep("bla ", 50), collapse = "")
  expect_identical(ids_to_codes(long_text), "SICI")
})

test_that("ids_download() works with single mocked item", {
  url <- "http://fake-url"

  local_mocked_bindings(
    req_perform = function(req) {
      structure(list(
        status_code = 200,
        body = '<?xml version="1.0"?><rss><channel><item>
                  <title>Test record</title>
                  <guid>MD_1234</guid>
                </item></channel></rss>'
      ), class = "response_httr2_mock")
    },
    resp_status = function(res) 200,
    resp_body_string = function(res) res$body,
    .package = "httr2"
  )

  df <- ids_download(url)

  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 1)
  expect_equal(df$id, "1234")
  expect_equal(df$title, "Test record")
  expect_equal(df$guid, "MD_1234")
})

test_that("ids_download() works with multiple mocked items", {
  url <- "http://fake-url"

  local_mocked_bindings(
    req_perform = function(req) {
      structure(list(
        status_code = 200,
        body = '<?xml version="1.0"?><rss><channel>
                  <item><title>Record 1</title><guid>MD_111</guid></item>
                  <item><title>Record 2</title><guid>MD_222</guid></item>
                  <item><title>Record 3</title><guid>MD_333</guid></item>
                </channel></rss>'
      ), class = "response_httr2_mock")
    },
    resp_status = function(res) 200,
    resp_body_string = function(res) res$body,
    .package = "httr2"
  )

  df <- ids_download(url)

  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 3)
  expect_equal(df$id, c("111", "222", "333"))
})

test_that("ids_download() returns empty dataframe when XML has no items", {
  url <- "http://fake-url"

  local_mocked_bindings(
    req_perform = function(req) {
      structure(list(
        status_code = 200,
        body = '<?xml version="1.0"?><rss><channel></channel></rss>'
      ), class = "response_httr2_mock")
    },
    resp_status = function(res) 200,
    resp_body_string = function(res) res$body,
    .package = "httr2"
  )

  df <- ids_download(url)|> suppressWarnings()

  expect_s3_class(df, "data.frame")
  expect_equal(nrow(df), 0)
})

test_that("ids_download() handles network errors gracefully", {
  url <- "http://fake-url"

  local_mocked_bindings(
    req_perform = function(req) stop("Network error"),
    .package = "httr2"
  )

  expect_error(suppressWarnings(ids_download(url)))
})

test_that("ids_url_build() builds a valid proxy URL", {
  extent <- list(left = 1, bottom = 2, right = 3, top = 4)
  url <- ids_url_build(extent, "75")

  expect_type(url, "character")
  expect_match(url, "^http://atlas\\.patrimoines\\.culture\\.fr/atlas/trunk/proxy_ign\\.php\\?url=")

  # Decode URL for readable matching
  decoded <- utils::URLdecode(url)
  expect_match(decoded, "geosource/srv/fr/rss\\.search")
  expect_match(decoded, "DEP_75")
  expect_match(decoded, "geoForm=BBOX")
})


test_that("ids_url_build() encodes special characters safely", {
  extent <- list(left = 0, bottom = 0, right = 1, top = 1)
  url <- ids_url_build(extent, "2A") # Corsican department

  expect_true(!grepl(" ", url))  # No spaces
  expect_true(grepl("DEP_2A", url))  # Properly encoded department code
})

test_that("ids_url_build() handles numeric department codes", {
  extent <- list(left = 0, bottom = 0, right = 1, top = 1)
  url <- ids_url_build(extent, 33)

  decoded <- utils::URLdecode(url)
  expect_match(decoded, "geoID=DEP_33")
})

test_that("ids_url_build() fails gracefully on malformed extent", {
  bad_extent <- list(left = 0, bottom = 0)  # Missing right/top
  expect_error(
    ids_url_build(bad_extent, "75"),
    regexp = "missing.*right|top",
    fixed = FALSE
  )
})

#' Retrieve heritage layer IDs for a given sf object
#'
#' This function retrieves available heritage dataset identifiers from the
#' French Ministry of Culture’s "Atlas du Patrimoine" service.
#'
#' It uses the spatial extent and inferred department(s) of the input `sf`
#' object to query the service.
#'
#' @param x An `sf` object defining the area of interest.
#' @param buffer A `numeric` value (default = 2500). Buffer distance in meters
#'   used to slightly expand geometries before querying.
#' @param crs An `integer` or `sf::st_crs` object (default = 2154). Coordinate
#'   reference system used for spatial processing.
#' @param verbose Logical. If `TRUE` (default), prints progress and diagnostic messages.
#'
#' @details
#' Internally, the function:
#' \enumerate{
#'   \item Aggregates nearby geometries using `buffer` input.
#'   \item Determines the corresponding INSEE department code for each geometry, using `happign::get_wfs()`.
#'   \item Computes the bounding box of each geometry.
#'   \item Queries the "Atlas du Patrimoine" service feed for all available metadata
#'         records (IDs, titles, GUIDs) within each bounding box.
#' }
#'
#' @return
#' A `data.frame` with the following columns:
#' \describe{
#'   \item{id}{Numeric identifier extracted from the record GUID.}
#'   \item{title}{Record title as published in the service feed.}
#'   \item{guid}{Full GUID (unique resource identifier).}
#'   \item{departement}{Department code associated with the spatial query.}
#'   \item{code}{Internal code associated with the layer.}
#' }
#' Returns an empty `data.frame` if no records are found or the request fails.
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'   # Create a minimal sf object defining the area of interest.
#'   # A simple point geometry is sufficient to trigger spatial queries.
#'   my_sf_layer <- sf::st_sf(
#'     geometry = sf::st_sfc(
#'       sf::st_point(c(2.21, 48.82)),
#'       crs = 4326
#'     )
#'   )
#'
#'   # Retrieve available heritage dataset identifiers intersecting
#'   # the spatial extent derived from the sf object.
#'   ids <- get_heritage_ids(my_sf_layer)
#' }
#' }
#'
#' @importFrom sf st_bbox
#'
#' @export
#'
get_heritage_ids <- function(x,
                             buffer = 2500,
                             crs = 2154,
                             verbose = TRUE) {

  # --- Step 0: Check service availability ----

  # Check whether the external Atlas service is available before any processing.
  # This avoids running expensive spatial operations if the backend is unreachable.
  if (!isTRUE(atlas_ok())) {
    stop("Atlas service is not available.", call. = FALSE)
  }

  # --- Step 1: Prepare + validate geometry ----

  # Prepare input geometry:
  # - reproject to target CRS
  # - apply buffer to extend spatial search area
  # - perform internal cleaning/normalisation of geometry
  y <- geo_prepare(x, crs = crs, buffer = buffer)

  # Validate that resulting object is a valid sf object with proper geometry
  geo_object_check(y)

  # Defensive check: ensure preprocessing did not remove all features
  if (is.null(y) || nrow(y) == 0L) {
    stop("Empty geometry after preprocessing.", call. = FALSE)
  }

  # --- Step 2.1: Compute extents ----

  # Compute bounding boxes (extent) for each feature.
  # These extents will be used to query the external spatial service.
  y_extents <- geo_extent(y)

  # Validate extent structure to ensure downstream query construction is safe
  if (!is.data.frame(y_extents) || nrow(y_extents) == 0L) {
    stop("Invalid extents.", call. = FALSE)
  }

  # Ensure expected extent schema is present (strict contract)
  required_cols <- c("left", "bottom", "right", "top")

  if (!all(required_cols %in% names(y_extents))) {
    stop("Invalid extent structure.", call. = FALSE)
  }

  y_extents <- y_extents[, required_cols, drop = FALSE]

  # --- Step 2.2: Compute departments (per geometry) ----

  # Extract individual geometries to compute administrative department membership
  # Each geometry is mapped to one or more department codes
  y_geoms <- sf::st_geometry(y)

  y_deps <- lapply(seq_along(y_geoms), function(i) {

    # Wrap single geometry into an sf object for spatial lookup
    quiet(get_deps(
      sf::st_sf(
        geometry = sf::st_sfc(y_geoms[[i]], crs = sf::st_crs(y))
      )
    ))
  })

  # Ensure spatial department mapping is consistent with input geometries
  if (length(y_deps) != length(y_geoms)) {
    stop("Internal error: mismatch between geometries and departments.", call. = FALSE)
  }

  # Fail fast if any geometry cannot be assigned to a department
  if (any(vapply(y_deps, function(d) length(d) == 0L, logical(1)))) {
    stop("At least one geometry has no department.", call. = FALSE)
  }

  # Ensure no missing values exist in department assignment results
  if (any(vapply(y_deps, function(d) any(is.na(d)), logical(1)))) {
    stop("NA detected in department assignment.", call. = FALSE)
  }

  # Optional progress information: number of department assignments generated
  if (verbose) {
    message(sum(lengths(y_deps)), " group dep(s) detected")
  }

  # --- Step 3: Build query table (extent × dep) ----

  # Build query input combining spatial extent and department filters
  # Each geometry can generate multiple department-based queries
  z <- Map(function(i, deps_y) {

    # Informational log per processing group (extent + departments)
    if (verbose) {
      message(sprintf("Processing group dep(s): %s",
                      paste(unique(deps_y), collapse = ", ")))
    }

    ext_y <- y_extents[i, , drop = FALSE]

    # Skip group if no valid departments exist (defensive guard)
    if (length(deps_y) == 0L) return(NULL)

    # Construct query table for API request generation
    data.frame(
      dep = deps_y,
      left = ext_y$left,
      bottom = ext_y$bottom,
      right = ext_y$right,
      top = ext_y$top,
      stringsAsFactors = FALSE
    )

  }, seq_along(y_deps), y_deps)

  # Remove empty groups resulting from filtering or invalid mappings
  z <- Filter(Negate(is.null), z)
  names(z) <- NULL

  # Flatten grouped structure into a single query table
  queries <- do.call(rbind, z)

  # Fail gracefully if no valid queries can be constructed
  if (is.null(queries) || nrow(queries) == 0L) {
    warning("No valid extent/department combinations.", call. = FALSE)
    return(data.frame())
  }

  # Build request URLs for each query row
  # Each URL encodes spatial extent + department filter
  queries$url <- vapply(seq_len(nrow(queries)), function(i) {

    extent <- list(
      left   = queries$left[i],
      bottom = queries$bottom[i],
      right  = queries$right[i],
      top    = queries$top[i]
    )

    ids_url_build(extent, queries$dep[i])

  }, character(1))

  # --- Step 4: Download ----

  # Execute API requests sequentially for each query
  results <- lapply(seq_len(nrow(queries)), function(i) {

    df <- ids_download(queries$url[i])

    # Handle empty or failed API responses gracefully
    if (!is.data.frame(df) || nrow(df) == 0L) {
      if (verbose) {
        message(sprintf("No result for dep %s", queries$dep[i]))
      }
      return(NULL)
    }

    # Attach department metadata to each returned record
    df$departement <- queries$dep[i]

    df
  })

  # --- Step 5: Combine ----

  # Keep only successful (non-empty) results
  valid <- Filter(function(x) is.data.frame(x) && nrow(x) > 0, results)

  # If all requests failed or returned empty results, exit safely
  if (length(valid) == 0L) {
    warning("No heritage identifiers were retrieved.", call. = FALSE)
    return(data.frame())
  }

  # Merge all partial results into a single data.frame
  combined <- do.call(rbind, valid)

  # Final safety check on merged output
  if (nrow(combined) == 0L) {
    warning("No heritage identifiers were retrieved.", call. = FALSE)
    return(data.frame())
  }

  # Ensure required columns exist before post-processing
  required_keys <- c("id", "departement")

  missing <- setdiff(required_keys, names(combined))
  if (length(missing) > 0L) {
    stop("Missing required columns in result: ", paste(missing, collapse = ", "), call. = FALSE)
  }

  # Remove duplicated records based on ID and department combination
  combined <- combined[!duplicated(combined[, required_keys, drop = FALSE]), ]

  # Derive internal classification code from title field
  combined$code <- ids_to_codes(combined$title)

  # Optional final summary message
  if (verbose) {
    message(sprintf("%d heritage ID(s) retrieved", nrow(combined)))
  }

  return(combined)
}

test_that("get_heritage_ids() returns combined heritage IDs", {

  x <- sf::st_as_sf(sf::st_sfc(
    sf::st_point(c(2.209271, 48.82095)), crs = 4326
  ))

  mocked_get_deps <- function(...) c("92")

  mocked_geo_extent <- function(...) {
    data.frame(
      left = 2.17522401805063,
      bottom = 48.7984671628133,
      right = 2.24331827597942,
      top = 48.8434323867205
    )
  }

  mocked_ids_url_build <- function(extent_row, dep) "http://fake-url"

  mocked_ids_download <- function(url, ...) {
    data.frame(
      id = c("4358", "3872", "3464"),
      title = c(
        "Registered Site - Ile-de-France (AC2)",
        "UNESCO World Heritage - Buffer zone footprint",
        "Classified buildings - Hauts-de-Seine - 92"
      ),
      guid = c("4358", "3872", "3464"),
      stringsAsFactors = FALSE
    )
  }

  mocked_ids_to_codes <- function(titles) {
    paste0("code_", seq_along(titles))
  }

  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = mocked_get_deps,
    geo_extent = mocked_geo_extent,
    ids_url_build = mocked_ids_url_build,
    ids_download = mocked_ids_download,
    ids_to_codes = mocked_ids_to_codes,
    .package = "frheritage"
  )

  res <- get_heritage_ids(x, verbose = FALSE)

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 3)
  expect_equal(res$id, c("4358", "3872", "3464"))
  expect_equal(res$guid, c("4358", "3872", "3464"))
  expect_equal(res$code, c("code_1", "code_2", "code_3"))
})

test_that("get_heritage_ids() handles errors gracefully", {

  x <- sf::st_as_sf(sf::st_sfc(
    sf::st_point(c(2.209271, 48.82095)), crs = 4326
  ))

  # Case 1: no department returned
  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) list(),  # important: empty
    geo_extent = function(...) data.frame(
      left = 1, bottom = 1, right = 2, top = 2
    ),
    .package = "frheritage"
  )

  expect_error(
    get_heritage_ids(x, verbose = FALSE),
    "At least one geometry has no department"
  )

  # Case 2: geo_extent returns NULL
  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) c("92"),
    geo_extent = function(...) NULL,
    .package = "frheritage"
  )

  expect_error(
    get_heritage_ids(x, verbose = FALSE),
    "Invalid extents\\."
  )

  # Case 3: empty download result
  local_mocked_bindings(
    atlas_ok = function() TRUE,
    get_deps = function(...) c("92"),
    geo_extent = function(...) data.frame(
      left = 1, bottom = 1, right = 2, top = 2
    ),
    ids_url_build = function(...) "http://fake-url",
    ids_download = function(...) data.frame(),
    ids_to_codes = function(...) character(0),
    .package = "frheritage"
  )

  expect_warning(
    res <- get_heritage_ids(x, verbose = FALSE),
    "No heritage identifiers were retrieved"
  )

  expect_s3_class(res, "data.frame")
  expect_equal(nrow(res), 0)
})

#' Retrieve and download heritage spatial data for a given sf object
#'
#' This function retrieves and downloads spatial heritage datasets from the
#' French Ministry of Culture’s "Atlas du Patrimoine" service.
#'
#' It uses the spatial extent and inferred department(s) of the input `sf`
#' object to query an internal registry of heritage datasets, then downloads
#' and merges the corresponding spatial layers.
#'
#' @param x An `sf` object defining the area of interest.
#' @param data_code A single `character` heritage dataset codes to retrieve.
#'   Valid codes can be obtained with [get_heritage_layernames()].
#' @param buffer A `numeric` value (default = 2500). Buffer distance in meters
#'   used to slightly expand geometries before querying.
#' @param crs An `integer` or `sf::st_crs` object (default = 2154). Coordinate
#'   reference system used for spatial processing.
#' @param verbose Logical. If `TRUE` (default), prints progress and diagnostic messages.
#'
#' @details
#' This function currently supports metropolitan France departments (96 units).
#'
#' Internally, the function:
#' \enumerate{
#'   \item Validates the input `data_code` against an internal registry.
#'   \item Prepares and optionally buffers the input geometry.
#'   \item Computes spatial extents (bounding boxes) for each feature.
#'   \item Assigns each feature to one or more administrative departments.
#'   \item Filters available heritage records using an internal dataset registry.
#'   \item Builds query URLs for the Atlas du Patrimoine service.
#'   \item Downloads ZIP archives containing shapefiles.
#'   \item Reads, validates, and merges the resulting `sf` objects.
#' }
#'
#' Partial failures are tolerated: invalid or empty responses for individual
#' requests are skipped.
#'
#' @return
#' A single `sf` object containing all retrieved heritage features.
#' Returns an empty `sf` object if no data is found or all requests fail.
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'
#'   # Define a minimal spatial object (WGS84 point)
#'   my_sf_layer <- sf::st_sf(
#'     geometry = sf::st_sfc(
#'       sf::st_point(c(2.21, 48.82)),
#'       crs = 4326
#'     )
#'   )
#'
#'   # Retrieve heritage spatial data
#'   res <- get_heritage(
#'     x = my_sf_layer,
#'     data_code = "IMMH",
#'     buffer = 2000,
#'     crs = 2154,
#'     verbose = TRUE
#'   )
#'
#' }
#' }
#'
#' @importFrom sf st_bbox st_crs
#'
#' @export
#'
get_heritage <- function(x,
                         data_code,
                         buffer = 2500,
                         crs = 2154,
                         verbose = TRUE) {

  # --- Step 0: Check service availability ----

  # Check whether the external Atlas service is available before any processing.
  # This avoids running expensive spatial operations if the backend is unreachable.
  if (!isTRUE(atlas_ok())) {
    stop("Atlas service is not available.", call. = FALSE)
  }

  # --- Step 1: Validate inputs ----

  # Validate that the requested heritage dataset code exists in the internal registry.
  data_check(data_code)

  # Enforce single dataset selection to avoid ambiguous multi-source queries.
  if (length(data_code) != 1L) {
    stop("`data_code` must be a single heritage code.", call. = FALSE)
  }

  # --- Step 2: Prepare + validate geometry ----

  # Prepare input geometry:
  # - reproject to target CRS
  # - apply buffer to extend spatial search area
  # - perform internal cleaning/normalisation of geometry
  y <- geo_prepare(x, crs = crs, buffer = buffer)
  geo_object_check(y)

  # Defensive check: ensure preprocessing did not remove all features
  if (is.null(y) || nrow(y) == 0L) {
    stop("Empty geometry after preprocessing.", call. = FALSE)
  }

  # --- Step 3.1: Compute extents ----

  # Compute bounding boxes (extent) for each feature.
  # These extents will be used to query the external spatial service.
  y_extents <- geo_extent(y)

  # Validate extent structure to ensure downstream query construction is safe
  if (!is.data.frame(y_extents) || nrow(y_extents) == 0L) {
    stop("Invalid extents.", call. = FALSE)
  }

  # Ensure expected extent schema is present (strict contract)
  required_cols <- c("left", "bottom", "right", "top")

  if (!all(required_cols %in% names(y_extents))) {
    stop("Invalid extent structure.", call. = FALSE)
  }

  y_extents <- y_extents[, required_cols, drop = FALSE]

  # --- Step 3.2: Compute departments ----

  # Extract individual geometries to compute administrative department membership
  # Each geometry is mapped to one or more department codes
  y_geoms <- sf::st_geometry(y)

  y_deps <- lapply(seq_along(y_geoms), function(i) {

    # Wrap single geometry into an sf object for spatial lookup
    quiet(get_deps(
      sf::st_sf(geometry = sf::st_sfc(y_geoms[[i]], crs = sf::st_crs(y)))
    ))
  })

  # Ensure spatial department mapping is consistent with input geometries
  if (length(y_deps) != length(y_geoms)) {
    stop("Internal error: mismatch between geometries and departments.", call. = FALSE)
  }

  # Fail fast if any geometry cannot be assigned to a department
  if (any(vapply(y_deps, function(d) length(d) == 0L, logical(1)))) {
    stop("At least one geometry has no department.", call. = FALSE)
  }

  # Ensure no missing values exist in department assignment results
  if (any(vapply(y_deps, function(d) any(is.na(d)), logical(1)))) {
    stop("NA detected in department assignment.", call. = FALSE)
  }

  # Optional progress information: number of department assignments generated
  if (verbose) {
    message(sum(lengths(y_deps)), " group dep(s) detected")
  }

  # --- Step 3.3: Build query groups (extent × dep × id) ----

  # Pre-split extents into indexed list for stable positional mapping.
  ext_list <- lapply(seq_len(nrow(y_extents)), function(i) {
    y_extents[i, , drop = FALSE]
  })

  z <- Map(function(i, deps_y) {

    # Extract spatial extent corresponding to current geometry index.
    ext_y <- ext_list[[i]]

    # Retrieve dataset records filtered by department and requested data code.
    df_ids <- data_filter(department = deps_y, data_code = data_code)

    # Skip group if no matching records exist in the dataset registry
    if (!is.data.frame(df_ids) || nrow(df_ids) == 0L) return(NULL)

    # Filter records strictly to requested departments (safety enforcement).
    df_ids <- df_ids[df_ids$departement %in% deps_y, , drop = FALSE]

    # Abort group if filtering removes all candidate records.
    if (nrow(df_ids) == 0L) {
      if (verbose) {
        message("No valid IDs after filtering for departments.")
      }
      return(NULL)
    }

    # Remove duplicate department entries to avoid redundant requests.
    df_ids <- df_ids[!duplicated(df_ids$departement), , drop = FALSE]

    if (nrow(df_ids) == 0L) return(NULL)

    # Construct final query table combining spatial extent and dataset metadata.
    data.frame(
      dep = df_ids$departement,
      id = df_ids$id,
      title = df_ids$title,
      guid = if ("guid" %in% names(df_ids)) df_ids$guid else NA_character_,
      left = rep(ext_y$left, nrow(df_ids)),
      bottom = rep(ext_y$bottom, nrow(df_ids)),
      right = rep(ext_y$right, nrow(df_ids)),
      top = rep(ext_y$top, nrow(df_ids)),
      stringsAsFactors = FALSE
    )

  }, seq_along(y_deps), y_deps)

  # Remove empty groups generated by filtering or invalid mappings.
  z <- Filter(Negate(is.null), z)
  names(z) <- NULL

  # --- Step 4: API requests ----

  # Internal helper function to validate that returned objects are usable sf objects before merging.
  is_valid_sf <- function(x) {

    if (is.null(x)) return(FALSE)
    if (!inherits(x, "sf")) return(FALSE)
    if (nrow(x) == 0L) return(FALSE)

    # Ensure geometry column exists and is not entirely empty.
    if (is.null(sf::st_geometry(x))) return(FALSE)
    if (all(sf::st_is_empty(x))) return(FALSE)

    TRUE
  }

  # Iterates over each query group (extent × department × dataset) and executes the full download and
  # spatial reconstruction workflow, returning a list of sf objects for each successfully processed group.
  results <- lapply(seq_along(z), function(k) {

    z_i <- z[[k]]

    if (is.null(z_i) || nrow(z_i) == 0L) return(NULL)

    if (verbose) {
      message("Processing group dep(s): ",
              paste(unique(z_i$dep), collapse = ", "))
    }

    # Internal helper function that processes a single query row by building the download request,
    # retrieving the corresponding spatial ZIP archive, and converting it into a validated sf object.
    process_row <- function(row) {

      # Build download request URL for zipped spatial dataset.
      url <- zip_query_build(
        id = row$id,
        title = row$title,
        guid = row$guid,
        extent_vals = c(row$left, row$bottom, row$right, row$top),
        crs = crs
      )

      # Download ZIP archive containing spatial data.
      zip_tmp <- zip_download(url, row$id)

      # Handle failed download attempts gracefully.
      if (is.null(zip_tmp)) {
        if (verbose) {
          message("Download failed for ID ", row$id,
                  " (dep ", row$dep, ")")
        }
        return(NULL)
      }

      # Extract and read shapefiles from downloaded archive.
      res <- geo_shapefile_read(zip_tmp, crs = crs)

      # Validate resulting spatial object integrity.
      if (!is_valid_sf(res)) {
        if (verbose) {
          message("Empty/invalid sf for ID ", row$id,
                  " (dep ", row$dep, ")")
        }
        return(NULL)
      }

      res
    }

    # Apply processing per dataset record in current group.
    res_i <- lapply(seq_len(nrow(z_i)), function(i) {
      process_row(z_i[i, ])
    })

    Filter(Negate(is.null), res_i)

  })

  # --- Step 5: Merge results ----

  flat <- unlist(results, recursive = FALSE)
  flat <- Filter(Negate(is.null), flat)

  # Return empty sf object if no valid data was retrieved.
  if (length(flat) == 0L) {
    return(
      sf::st_sf(
        geometry = sf::st_sfc(crs = crs)
      )[0, ]
    )
  }

  # Merge list of sf objects into a single spatial dataset.
  merged <- geo_sf_bind(flat)

  # Remove duplicated geometries to ensure dataset uniqueness.
  if (!is.null(merged) && nrow(merged) > 0L) {
    merged <- merged[!duplicated(sf::st_geometry(merged)), ]
  }

  # Optional summary of retrieved features.
  if (verbose) {
    message(sprintf("%d feature(s) retrieved", nrow(merged)))
  }

  return(merged)
}

# geometry type section ----
#' Determine the geometry type of an `sf` object
#'
#' This internal helper function identifies and normalizes the geometry type
#' of an `sf` object. It ensures that all geometries within the object share
#' a consistent type (e.g., `"POINT"`, `"LINESTRING"`, `"POLYGON"`).
#' If multiple geometry types are found, an error is raised.
#'
#' @param x An `sf` object whose geometry type should be determined.
#'
#' @details
#' The function retrieves the geometry type of the input using
#' [sf::st_geometry_type()] and standardizes it according to a simplified
#' classification:
#' \itemize{
#'   \item `"POINT"` and `"MULTIPOINT"` = `"POINT"`
#'   \item `"LINESTRING"` and `"MULTILINESTRING"` = `"LINE"`
#'   \item `"POLYGON"` and `"MULTIPOLYGON"` = `"POLYGON"`
#'   \item All other types = `"GEOMETRY"`
#' }
#'
#' Mixed or inconsistent geometry collections are not supported.
#'
#' @return
#' A single `character(1)` value representing the normalized geometry type.
#' Possible values are: `"POINT"`, `"LINE"`, `"POLYGON"`, or `"GEOMETRY"`.
#'
#' @importFrom sf st_geometry_type
#'
#' @keywords internal
#'
geo_object_type <- function(x) {
  a <- list(
    POINT = "POINT",
    POINT = "MULTIPOINT",
    LINE = "LINESTRING",
    LINE = "MULTILINESTRING",
    POLYGON = "POLYGON",
    POLYGON = "MULTIPOLYGON",
    other = "GEOMETRY",
    other = "GEOMETRYCOLLECTION",
    other = "CIRCULARSTRING",
    other = "COMPOUNDCURVE",
    other = "CURVEPOLYGON",
    other = "MULTICURVE",
    other = "MULTISURFACE",
    other = "CURVE",
    other = "SURFACE",
    other = "POLYHEDRALSURFACE",
    other = "TIN",
    other = "TRIANGLE"
  )
  type <- sf::st_geometry_type(x)
  levels(type) <- a
  type <- as.character(unique(type))
  if (length(type) > 1) {
    stop("GEOMETRYCOLLECTION objects should have consistent type",
         call. = FALSE
    )
  }
  return(type)
}

#' Check that spatial objects are valid `sf` layers with expected geometry type
#'
#' This internal helper verifies that all provided objects:
#' \itemize{
#'   \item are of class `sf` (optional, controlled by `check_class`);
#'   \item contain valid geometries (`sf::st_is_valid()`);
#'   \item contain no empty geometries (`!sf::st_is_empty()`);
#'   \item have a geometry type among the allowed ones (if specified).
#' }
#'
#' @param ... One or several spatial objects to test.
#' @param check_class Logical. Whether to check that each object inherits from
#' class `sf`. Default is `TRUE`.
#' @param check_valid Logical. Whether to check that geometries are valid.
#' Default is `TRUE`.
#' @param check_empty Logical. Whether to check that geometries are not empty.
#' Default is `TRUE`.
#' @param allowed_geom_classes Character vector of allowed geometry types
#' (e.g., `"POLYGON"`, `"LINE"`, `"POINT"`). If `NULL`, the geometry type
#' is not checked. Default is `"POLYGON"`.
#'
#' @details
#' The function raises an informative error message if any condition fails.
#' It is intended for internal validation of `sf` inputs before performing
#' geometric operations or API queries.
#'
#' Geometry type validation is performed via [geo_object_type()], which ensures
#' type consistency across geometries.
#'
#' @return
#' Invisibly returns `NULL`. Throws an error if any object fails validation.
#'
#' @seealso [geo_object_type()]
#'
#' @importFrom sf st_is_valid st_is_empty
#'
#' @keywords internal
#'
geo_object_check <- function(...,
                             check_class = TRUE,
                             check_valid = TRUE,
                             check_empty = TRUE,
                             allowed_geom_classes = "POLYGON") {

  dots <- list(...)
  arg_names <- as.list(substitute(list(...)))[-1L]

  for (i in seq_along(dots)) {
    obj <- dots[[i]]
    name <- deparse(arg_names[[i]])

    # Check class
    if (check_class && !inherits(obj, "sf")) {
      stop(sprintf("Object '%s' must be of class 'sf'.", name), call. = FALSE)
    }

    # Check geometry validity
    if (check_valid && any(!sf::st_is_valid(obj))) {
      stop(sprintf("Object '%s' contains invalid geometries.", name), call. = FALSE)
    }

    # Check for empty geometries
    if (check_empty && any(sf::st_is_empty(obj))) {
      stop(sprintf("Object '%s' contains empty geometries.", name), call. = FALSE)
    }

    # Check allowed geometry types
    if (!is.null(allowed_geom_classes)) {
      geom_type <- geo_object_type(obj)
      if (!geom_type %in% allowed_geom_classes) {
        stop(sprintf(
          "Object '%s' must have geometry type: %s (found: '%s').",
          name,
          paste(allowed_geom_classes, collapse = ", "),
          geom_type
        ), call. = FALSE)
      }
    }
  }

  invisible(NULL)
}

# sf section ----
#' Prepare and aggregate an sf object
#'
#' This helper function prepares and aggregates an `sf` object
#' to produce clean, valid, and unified geometries.
#'
#' @param x An `sf` object to process.
#' @param crs Integer. Code of the target CRS. Default is `2154`.
#' @param buffer Numeric. Buffer distance (in map units) used to merge nearby features. Default is `10`.
#'
#' @return An `sf` object with cleaned, projected, and aggregated geometries as `POLYGON`s.
#'
#' @details
#' The function performs a complete geometric preparation workflow:
#'
#' 1. Geometry validation: ensures that all geometries belong to allowed classes (`POINT`, `LINE`, or `POLYGON`).
#' 2. Geometry cleaning: invalid features are fixed with [sf::st_make_valid()], Z/M dimensions are dropped via [sf::st_zm()], and multipart geometries are cast to simple forms.
#' 3. Coordinate transformation: all geometries are transformed to the specified CRS using [sf::st_transform()].
#' 4. Aggregation: small buffer zones (controlled by `buffer`) are applied to merge adjacent or overlapping features.
#'    The buffered geometries are then dissolved with [sf::st_union()] and cast to `POLYGON` type.
#'
#' This process is useful for simplifying feature sets into larger contiguous study zones or analysis areas.
#'
#' @importFrom sf st_make_valid st_zm st_transform st_buffer st_union st_cast st_sf
#'
#' @keywords internal
#'
geo_prepare <- function(x, crs = 2154, buffer = 10) {
  # Validate geometry classes
  geo_object_check(x, allowed_geom_classes = c("POINT", "LINE", "POLYGON"))

  # Clean geometries
  x <- quiet(st_make_valid(x))
  x <- quiet(st_zm(x))
  x <- quiet(geo_cast(x))
  x <- st_transform(x, crs)

  # Aggregation
  buffer <- max(10, buffer)
  x <- st_buffer(x, dist = buffer)
  x <- st_union(x)
  x <- st_sf(st_cast(x, "POLYGON", warn = FALSE))
  x <- quiet(st_make_valid(x))
  x
}

#' Compute bounding boxes for each feature in an sf object
#'
#' Computes one bounding box per geometry in an `sf` object.
#'
#' @param x An `sf` object (POLYGON or MULTIPOLYGON) for which to compute bounding boxes.
#' @param crs Numeric or sf CRS. Target CRS used before computation (default 4326).
#'
#' @return A `data.frame` with one row per feature:
#' \describe{
#'   \item{left}{xmin}
#'   \item{bottom}{ymin}
#'   \item{right}{xmax}
#'   \item{top}{ymax}
#' }
#'
#' @details
#' - The function first validates the input with [geo_object_check()].
#' - Transforms the object to the target CRS before computing bounding boxes.
#' - CRS and units are inherited from the input `sf` object.
#'
#' @importFrom sf st_transform st_geometry st_bbox
#'
#' @keywords internal
#'
geo_extent <- function(x, crs = 4326) {

  geo_object_check(x)

  x <- sf::st_transform(x, crs)

  geoms <- sf::st_geometry(x)

  if (length(geoms) == 0L) {
    return(
      data.frame(
        left = numeric(0),
        bottom = numeric(0),
        right = numeric(0),
        top = numeric(0)
      )
    )
  }

  res <- lapply(geoms, function(g) {
    bb <- sf::st_bbox(g)

    data.frame(
      left   = unname(bb["xmin"]),
      bottom = unname(bb["ymin"]),
      right  = unname(bb["xmax"]),
      top    = unname(bb["ymax"])
    )
  })

  do.call(rbind, res)
}

#' Cast geometries to simple types
#'
#' This internal helper converts multi-part geometries to their single-part
#' equivalents: MULTIPOLYGON became POLYGON, MULTILINESTRING became LINESTRING,
#' MULTIPOINT became POINT. If the input contains mixed types, each feature is
#' cast individually.
#'
#' @param x An `sf` object with geometries to cast.
#'
#' @return An `sf` object with geometries cast to simple types.
#'
#' @importFrom sf st_geometry_type st_cast
#'
#' @keywords internal
#'
geo_cast <- function(x) {
  geo_object_check(x, allowed_geom_classes = c("POINT", "LINE", "POLYGON"))

  geom_types <- as.character(sf::st_geometry_type(x))

  # Mapping multi-part geometries to simple equivalents
  single_map <- c(
    MULTIPOLYGON    = "POLYGON",
    MULTILINESTRING = "LINESTRING",
    MULTIPOINT      = "POINT"
  )

  # Helper to cast a single feature if needed
  cast_one <- function(obj, geom_type) {
    if (geom_type %in% names(single_map)) {
      sf::st_cast(obj, single_map[[geom_type]])
    } else {
      obj
    }
  }

  # If all geometries are of the same type, cast all at once
  unique_types <- unique(geom_types)
  if (length(unique_types) == 1) {
    type <- unique_types
    if (type %in% names(single_map)) {
      x <- sf::st_cast(x, single_map[[type]])
    }
    return(x)
  }

  # Otherwise, cast feature by feature
  x_list <- vector("list", nrow(x))
  for (i in seq_len(nrow(x))) {
    x_list[[i]] <- cast_one(x[i, ], geom_types[i])
  }

  do.call(rbind, x_list)
}

#' Read and transform shapefile from a ZIP archive
#'
#' Extracts shapefiles from a ZIP archive, reads the shapefile found
#' in the "documents" folder into an `sf` object, and transforms it to
#' CRS:2154 (Lambert-93 projection for France), unless already in the target CRS.
#'
#' @param zip_path `character`. Path to the ZIP archive to read.
#' @param crs `numeric` or `sf::st_crs` object. Target CRS to transform to.
#' Default is 2154 (Lambert-93).
#'
#' @return An `sf` object transformed to CRS:2154.
#'
#' @importFrom sf st_read st_crs st_set_crs st_transform
#' @importFrom utils unzip
#'
#' @keywords internal
#'
geo_shapefile_read <- function(zip_path, crs = 2154) {
  # Extract ZIP to temporary directory
  exdir <- tempfile()
  utils::unzip(zip_path, exdir = exdir)

  # Find documents folder (could be at root or inside subdirectories)
  all_dirs <- list.dirs(exdir, recursive = TRUE, full.names = TRUE)
  documents_dir <- all_dirs[grepl("/documents$|\\\\documents$", all_dirs)]

  # Check if documents folder exists
  if (length(documents_dir) == 0) {
    stop("The 'documents' folder does not exist in the ZIP archive")
  }
  documents_dir <- documents_dir[1]

  if (!dir.exists(documents_dir)) {
    stop("The 'documents' folder does not exist in the ZIP archive")
  }

  # Find all shapefiles in documents folder
  shp_files <- list.files(documents_dir, pattern = "\\.shp$", full.names = TRUE, recursive = FALSE)

  # Check if there's exactly one shapefile
  if (length(shp_files) == 0) {
    stop("No shapefile (.shp) found in the 'documents' folder")
  }
  if (length(shp_files) > 1) {
    stop("Multiple shapefiles found in the 'documents' folder. Expected exactly one.")
  }

  # Read the shapefile
  sf_obj <- sf::st_read(shp_files[1], quiet = TRUE)

  # Transform to target CRS if needed
  if (is.na(sf::st_crs(sf_obj)$epsg) || sf::st_crs(sf_obj)$epsg != crs) {
    sf_obj <- sf::st_transform(sf_obj, crs)
  }

  return(sf_obj)
}


#' Bind multiple sf objects into a single sf
#'
#' This internal helper takes a list of `sf` objects, aligns their columns,
#' and binds them together into a single `sf` object. It removes NULL or empty
#' objects and handles nested lists of `sf`.
#'
#' @param sf_list A list of `sf` objects (can include nested lists).
#'
#' @return A single `sf` object containing all input features, or `NULL`
#' if the input list is empty or contains no valid sf objects.
#'
#' @keywords internal
#'
geo_sf_bind <- function(sf_list) {
  # Handle empty input
  if (length(sf_list) == 0) return(NULL)

  # If the first element is itself a list of sf, flatten one level
  if (inherits(sf_list[[1]], "list") && all(vapply(sf_list[[1]], inherits, logical(1), "sf"))) {
    sf_list <- sf_list[[1]]
  }

  # Remove NULL or empty sf
  sf_list <- Filter(function(x) inherits(x, "sf") && nrow(x) > 0, sf_list)
  if (length(sf_list) == 0) return(NULL)

  # Align columns
  ref_cols <- unique(unlist(lapply(sf_list, colnames)))
  aligned <- lapply(sf_list, function(sf_obj) {
    missing <- setdiff(ref_cols, colnames(sf_obj))
    for (c in missing) sf_obj[[c]] <- NA
    sf_obj[, ref_cols, drop = FALSE]
  })

  # Bind all sf together
  if (length(aligned) == 1) aligned[[1]] else do.call(rbind, aligned)
}

test_that("geo_object_check() validates sf objects correctly", {
  skip_if_not_installed("sf")

  # Create valid geometries
  point <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(1, 1))))
  multipoint <- sf::st_sf(geometry = sf::st_sfc(sf::st_multipoint(matrix(c(1,1,2,2), ncol = 2, byrow = TRUE))))
  line <- sf::st_sf(geometry = sf::st_sfc(sf::st_linestring(matrix(c(1,1,2,2), ncol = 2, byrow = TRUE))))
  multiline <- sf::st_sf(geometry = sf::st_sfc(sf::st_multilinestring(list(matrix(c(1,1,2,2), ncol = 2, byrow = TRUE)))))
  polygon <- sf::st_sf(geometry = sf::st_sfc(sf::st_polygon(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE)))))
  multipolygon <- sf::st_sf(geometry = sf::st_sfc(sf::st_multipolygon(list(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE))))))

  # Valid cases: should not raise errors
  expect_silent(geo_object_check(point, allowed_geom_classes = "POINT"))
  expect_silent(geo_object_check(multipoint, allowed_geom_classes = "POINT"))
  expect_silent(geo_object_check(line, allowed_geom_classes = "LINE"))
  expect_silent(geo_object_check(multiline, allowed_geom_classes = "LINE"))
  expect_silent(geo_object_check(polygon, allowed_geom_classes = "POLYGON"))
  expect_silent(geo_object_check(multipolygon, allowed_geom_classes = "POLYGON"))

  # Invalid class: not an sf object
  not_sf <- data.frame(x = 1:2, y = 3:4)
  expect_error(geo_object_check(not_sf), "must be of class 'sf'")

  # Invalid geometry type
  expect_error(
    geo_object_check(point, allowed_geom_classes = "POLYGON"),
    regexp = "must have geometry type"
  )

  # Invalid geometries
  invalid_poly <- sf::st_sf(
    geometry = sf::st_sfc(
      sf::st_polygon(list(matrix(c(0,0, 1,1, 1,0, 0,1, 0,0), ncol = 2, byrow = TRUE)))
    )
  )
  expect_error(
    geo_object_check(invalid_poly),
    "contains invalid geometries"
  )

  # Empty geometries
  empty_geom <- sf::st_sf(geometry = sf::st_sfc(sf::st_geometrycollection()))
  expect_error(geo_object_check(empty_geom), "contains empty geometries")
})

# Prepare sample geometries
point <- sf::st_sf(geometry = sf::st_sfc(sf::st_point(c(0,0))))
multipoint <- sf::st_sf(geometry = sf::st_sfc(sf::st_multipoint(matrix(c(0,0,1,1), ncol = 2, byrow = TRUE))))
line <- sf::st_sf(geometry = sf::st_sfc(sf::st_linestring(matrix(c(0,0,1,1), ncol = 2, byrow = TRUE))))
multiline <- sf::st_sf(geometry = sf::st_sfc(sf::st_multilinestring(list(matrix(c(0,0,1,1), ncol = 2, byrow = TRUE)))))
polygon <- sf::st_sf(geometry = sf::st_sfc(sf::st_polygon(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE)))))
multipolygon <- sf::st_sf(geometry = sf::st_sfc(sf::st_multipolygon(list(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE))))))

test_that("geo_cast() converts multi-part geometries to simple types", {
  expect_identical(as.character(sf::st_geometry_type(geo_cast(multipoint))), c("POINT", "POINT"))
  expect_identical(as.character(sf::st_geometry_type(geo_cast(multiline))), "LINESTRING")
  expect_identical(as.character(sf::st_geometry_type(geo_cast(multipolygon))), "POLYGON")
})

test_that("geo_cast() leaves single-part geometries unchanged", {
  expect_identical(as.character(sf::st_geometry_type(geo_cast(point))), "POINT")
  expect_identical(as.character(sf::st_geometry_type(geo_cast(line))), "LINESTRING")
  expect_identical(as.character(sf::st_geometry_type(geo_cast(polygon))), "POLYGON")
})

test_that("geo_cast() handles mixed geometry types feature by feature", {
  mixed_points <- rbind(point, multipoint)
  res <- geo_cast(mixed_points)
  geom_types <- as.character(st_geometry_type(res))
  expect_true(all(geom_types %in% "POINT"))

  mixed_polys <- rbind(polygon, multipolygon)
  res2 <- geo_cast(mixed_polys)
  geom_types2 <- as.character(st_geometry_type(res2))
  expect_true(all(geom_types2 %in% "POLYGON"))
})

test_that("geo_cast() runs silently on multiple geometries", {
  mixed_points <- rbind(point, multipoint)
  res <- geo_cast(mixed_points)
  geom_types <- as.character(sf::st_geometry_type(res))
  expect_true(all(geom_types %in% "POINT"))
})

test_that("geo_sf_bind() combines sf objects correctly", {
  skip_if_not_installed("sf")

  # Prepare minimal test geometries
  sf1 <- sf::st_sf(a = 1:2, geometry = sf::st_sfc(
    sf::st_point(c(0,0)), sf::st_point(c(1,1))
  ))
  sf2 <- sf::st_sf(a = 3:4, geometry = sf::st_sfc(
    sf::st_point(c(2,2)), sf::st_point(c(3,3))
  ))
  sf3 <- sf::st_sf(b = 5:6, geometry = sf::st_sfc(
    sf::st_point(c(4,4)), sf::st_point(c(5,5))
  ))
  empty_sf <- sf::st_sf(a = integer(0), geometry = sf::st_sfc())

  # Test: simple bind
  res <- geo_sf_bind(list(sf1, sf2))
  expect_s3_class(res, "sf")
  expect_equal(nrow(res), 4)
  expect_true(all(c("a", "geometry") %in% colnames(res)))

  # Test: bind with differing columns
  res2 <- geo_sf_bind(list(sf1, sf3))
  expect_s3_class(res2, "sf")
  expect_equal(nrow(res2), 4)
  expect_true(all(c("a", "b", "geometry") %in% colnames(res2)))
  expect_true(all(is.na(res2$b[1:2])))  # missing values filled with NA
  expect_true(all(is.na(res2$a[3:4])))

  # Test: nested list of sf
  nested <- list(list(sf1, sf2))
  res3 <- geo_sf_bind(nested)
  expect_s3_class(res3, "sf")
  expect_equal(nrow(res3), 4)

  # Test: list with NULL or empty sf
  res4 <- geo_sf_bind(list(sf1, NULL, empty_sf, sf2))
  expect_s3_class(res4, "sf")
  expect_equal(nrow(res4), 4)

  # Test: empty input returns NULL
  expect_null(geo_sf_bind(list()))
})

test_that("geo_object_type() correctly identifies geometry types", {
  point <- sf::st_sfc(sf::st_point(c(1, 1)))
  multipoint <- sf::st_sfc(sf::st_multipoint(matrix(c(1, 1, 2, 2), ncol = 2, byrow = TRUE)))
  line <- sf::st_sfc(sf::st_linestring(matrix(c(1, 1, 2, 2), ncol = 2, byrow = TRUE)))
  multiline <- sf::st_sfc(sf::st_multilinestring(list(matrix(c(1, 1, 2, 2), ncol = 2, byrow = TRUE))))
  polygon <- sf::st_sfc(sf::st_polygon(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE))))
  multipolygon <- sf::st_sfc(sf::st_multipolygon(list(list(matrix(c(0,0,1,0,1,1,0,1,0,0), ncol = 2, byrow = TRUE)))))

  expect_identical(geo_object_type(point), "POINT")
  expect_identical(geo_object_type(multipoint), "POINT")       # mapped to "POINT"
  expect_identical(geo_object_type(line), "LINE")              # mapped to "LINE"
  expect_identical(geo_object_type(multiline), "LINE")         # mapped to "LINE"
  expect_identical(geo_object_type(polygon), "POLYGON")
  expect_identical(geo_object_type(multipolygon), "POLYGON")   # mapped to "POLYGON"
})

#' Check that layer data codes are valid
#'
#' Validates that provided data codes exist in the heritage layers.
#'
#' @param data_code `character` vector. Data codes to check.
#'
#' @return `NULL`. Invisibly returns `NULL` if all codes are valid; otherwise throws an error.
#'
#' @details
#' - Uses `get_heritage_layernames()` to get the list of valid codes.
#' - Throws an error if any `data_code` is invalid.
#'
#' @seealso [get_heritage_layernames()]
#'
#' @keywords internal
#'
data_check <- function(data_code) {
  valid_codes <- get_heritage_layernames()$code
  if (!all(data_code %in% valid_codes)) {
    stop("`data_code` must be one of: ", paste(valid_codes, collapse = ", "),
         call. = FALSE)
  }
  invisible(NULL)
}

#' Filter layer IDs by internal code and department
#'
#' Filters a dataset of layer IDs based on provided data codes and departments.
#'
#' @param ids `data.frame`. Input dataset of IDs. `NULL` as default.
#' If `NULL` function uses `frheritage::all_ids`.
#' @param department `character` vector. Department(s) to filter.
#' @param data_code `character` vector. Data codes to filter. Must be valid heritage layer codes.
#'
#' @return `data.frame`. Subset of `ids` filtered by `data_code` and `department`, with `departement` column removed.
#'
#' @details
#' - Uses `data_check()` to ensure that all `data_code` values are valid.
#' - The function currently ignores the `ids` argument and always uses `frheritage::all_ids`.
#'
#' @keywords internal
#'
data_filter <- function(ids = NULL, department, data_code) {
  data_check(data_code)
  if (is.null(ids)){
    ids <- frheritage::all_ids
  }
  ids[ids[["code"]] %in% data_code & ids[["departement"]] %in% department, , drop = FALSE]
}

test_that("data_filter() returns subset with valid inputs", {
  skip_if_not_exists <- function() skip("frheritage::all_ids not available for testing")
  if (!exists("all_ids", envir = asNamespace("frheritage"))) skip_if_not_exists()

  # Pick a sample code and department from frheritage::all_ids
  sample_row <- frheritage::all_ids[1, , drop = FALSE]
  code_val <- sample_row$code
  dept_val <- sample_row$departement

  res <- data_filter(department = dept_val, data_code = code_val)
  expect_s3_class(res, "data.frame")
  expect_true(all(res$code %in% code_val))
})

test_that("data_filter() errors on invalid data_code", {
  expect_error(
    data_filter(department = "75", data_code = "INVALID"),
    regexp = "`data_code` must be one of"
  )
})

test_that("data_filter() uses all_ids if ids is NULL", {
  sample_row <- frheritage::all_ids[1, , drop = FALSE]
  code_val <- sample_row$code
  dept_val <- sample_row$departement

  res <- data_filter(ids = NULL, department = dept_val, data_code = code_val)
  expect_s3_class(res, "data.frame")
})

test_that("data_check() passes with valid codes", {
  expect_silent(data_check("IMDN"))
  expect_silent(data_check(c("IMDN", "IMMH")))
})

test_that("data_check() errors on a single invalid code", {
  expect_error(
    data_check("INVALID"),
    regexp = "`data_code` must be one of"
  )
})

test_that("data_check() errors when part of codes are invalid", {
  expect_error(
    data_check(c("IMDN", "INVALID")),
    regexp = "`data_code` must be one of"
  )
})

test_that("data_check() invisibly returns NULL for valid input", {
  res <- data_check("IMDN")
  expect_null(res)
})

#' Retrieve INSEE department codes from an sf object
#'
#' This function retrieves the INSEE administrative department codes
#' (`code_insee`) intersecting the features of a given `sf` object.
#' It relies on the *Admin Express COG* WFS service provided by IGN,
#' accessed via the **happign** package.
#'
#' @param x An `sf` object defining the area(s) of interest.
#'
#' @return A `character` vector of unique INSEE department codes intersecting
#' the input geometries. Returns `NULL` if no intersection is found or if the
#' WFS request fails.
#'
#' @details
#' The function performs the following steps:
#' \enumerate{
#'   \item Transforms the input geometries to CRS:4326.
#'   \item Computes centroids of the geometries.
#'   \item Queries the IGN WFS service (*Admin Express COG*) to retrieve departments.
#'   \item Performs a spatial join using [sf::st_intersects()].
#'   \item Extracts and returns unique `code_insee` values.
#' }
#'
#' Notes:
#' \itemize{
#'   \item The function depends on the availability of the IGN WFS service.
#'   \item Large or complex geometries may cause the request to fail.
#'   \item In such cases, consider simplifying or reducing the spatial extent of `x`.
#' }
#'
#' @examples
#' \donttest{
#' if (interactive()) {
#'
#'   # Minimal sf object (point in Paris)
#'   x <- sf::st_sf(
#'     geometry = sf::st_sfc(
#'       sf::st_point(c(2.35, 48.85)),
#'       crs = 4326
#'     )
#'   )
#'
#'   # Retrieve department code
#'   get_deps(x)
#'
#'   # Multiple locations (Paris and Rouen)
#'   x2 <- sf::st_sf(
#'     geometry = sf::st_sfc(
#'       sf::st_point(c(2.35, 48.85)),  # 75
#'       sf::st_point(c(1.09, 49.44)),  # 76
#'       crs = 4326
#'     )
#'   )
#'
#'   get_deps(x2)
#'
#' }
#' }
#'
#' @importFrom sf st_transform st_centroid st_geometry st_sf st_join st_intersects
#' @importFrom happign get_wfs intersects
#'
#' @export
get_deps <- function(x) {
  # Coordinate transformation to CRS:4326
  x <- sf::st_transform(x, 4326)
  x <- quiet(sf::st_centroid(x))

  # Try to get departments safely
  dep <- tryCatch({
    happign::get_wfs(
      x = x,
      layer = "ADMINEXPRESS-COG-CARTO.LATEST:departement",
      predicate = happign::intersects()
    )
  }, error = function(e) {
    message("happign::get_wfs() failed: Please try to reduce the spatial size")
    return(NULL)
  })

  # If no departments were retrieved, return NA vector
  if (is.null(dep) || nrow(dep) == 0) {
    message("No departments intersect x: Please try to reduce the spatial size")
    return(NULL)
  }

  # Join spatially to get INSEE codes
  joined <- sf::st_join(sf::st_sf(sf::st_geometry(x)),
                        dep[, "code_insee", drop = FALSE],
                        join = sf::st_intersects, left = TRUE)
  unique(joined$code_insee)
}

#' Build a URL query to download a ZIP
#'
#' Constructs a URL query to request a ZIP file containing spatial data from
#' the French Ministry of Culture's "Atlas du Patrimoine" service.
#'
#' @param id `character` or `numeric`. Identifier of the dataset.
#' @param title `character`. Title of the dataset.
#' @param guid `character`, optional. GUID of the dataset, if available.
#' @param extent_vals `numeric` vector of length 4. Bounding box coordinates in
#' order: left, bottom, right, top.
#' @param crs `numeric`, default `2154`. CRS code for the coordinate reference
#' system to use in the query.
#'
#' @return `character`. URL string for downloading the dataset in ZIP format.
#'
#' @details
#' - The function formats the bounding box and metadata into a JSON object,
#'   then encodes it into a URL suitable for the Atlas Patrimoines export service.
#' - CRS is converted to a string `"EPSG:<code>"`.
#'
#' @importFrom jsonlite toJSON
#' @importFrom sf st_crs
#'
#' @keywords internal
#'
zip_query_build <- function(id, title, guid = NULL, extent_vals, crs = 2154) {
  extent_vals <- unname(as.numeric(extent_vals))
  if (length(extent_vals) != 4) stop("extent_vals must have 4 numeric values: c(left, bottom, right, top)")

  # extent_vals = c(left, bottom, right, top)
  extent_list <- list(
    left   = extent_vals[1],
    bottom = extent_vals[2],
    right  = extent_vals[3],
    top    = extent_vals[4]
  )

  # Prepare metadata
  meta <- list(
    id = as.character(id),
    title = title,
    url = ""
  )
  if (!is.null(guid)) meta$guid <- guid

  # Build query
  query <- list(
    extent = extent_list,
    format = "SHP",
    srs = paste0("EPSG:", sf::st_crs(crs)$epsg),
    md = list(meta)
  )

  # Convert to JSON
  json_param <- jsonlite::toJSON(query, auto_unbox = TRUE)
  json_param <- gsub('\\"', "'", json_param, fixed = TRUE)

  # URL encode
  paste0(
    "http://atlas.patrimoines.culture.fr/services/export.php?data=",
    URLencode(json_param, reserved = TRUE)
  )
}

#' Download a ZIP file from a given URL
#'
#' Downloads a ZIP file from a specified URL and saves it to a temporary file.
#'
#' @param url `character`. URL to download the ZIP file from.
#' @param id `character` or `numeric`. Identifier of the dataset (used for messages).
#'
#' @return `character` or `NULL`.
#' Path to the temporary ZIP file, or `NULL` if download fails.
#'
#' @details
#' - Uses `httr2` to perform the HTTP request.
#' - If the HTTP response status is not 200, a warning is emitted and `NULL` is returned.
#' - Errors during download are caught and a warning is printed.
#'
#' @importFrom httr2 request req_perform resp_status resp_body_raw
#'
#' @keywords internal
#'
zip_download <- function(url, id) {
  tryCatch({
    resp <- request(url) |> req_perform()

    if (resp_status(resp) != 200) {
      warning(
        sprintf("HTTP %s for ID %s", resp_status(resp), id),
        call. = FALSE
      )
      return(NULL)
    }

    tmpf <- tempfile(fileext = ".zip")
    writeBin(resp_body_raw(resp), tmpf)
    tmpf

  }, error = function(e) {
    warning(
      sprintf("Download error for ID %s: %s", id, e$message),
      call. = FALSE
    )
    invisible(NULL)
  })
}

test_that("zip_query_build() builds a valid URL with minimal inputs", {
  id <- 1234
  title <- "Test Title"
  extent_vals <- c(100, 200, 300, 400)

  url <- zip_query_build(id, title, extent_vals = extent_vals, crs = 2154)

  expect_type(url, "character")
  expect_true(grepl("^http://atlas.patrimoines.culture.fr/services/export.php\\?data=", url))
  expect_true(grepl("EPSG%3A2154", url))
  expect_true(grepl("Test Title", URLdecode(url)))
  expect_true(grepl("123", URLdecode(url)))
})

test_that("zip_query_build() includes guid when provided", {
  id <- 4567
  title <- "Another Test"
  guid <- "MD_456"
  extent_vals <- c(10, 20, 30, 40)

  url <- zip_query_build(id, title, guid = guid, extent_vals = extent_vals)

  decoded <- URLdecode(url)
  expect_true(grepl("MD_456", decoded))
  expect_true(grepl("Another Test", decoded))
})

test_that("zip_query_build() fails with invalid extent length", {
  id <- 1
  title <- "Fail Case"
  extent_vals <- c(1,2,3)  # only 3 values

  expect_error(
    zip_query_build(id, title, extent_vals = extent_vals),
    "extent_vals must have 4 numeric values"
  )
})

test_that("zip_query_build() works with different CRS", {
  id <- 6789
  title <- "CRS Test"
  extent_vals <- c(0, 0, 1, 1)

  # Using a common geographic CRS (EPSG:4326)
  url <- zip_query_build(id, title, extent_vals = extent_vals, crs = 4326)

  decoded <- URLdecode(url)
  expect_true(grepl("EPSG:4326", decoded))
})

test_that("zip_query_build() JSON format uses single quotes after gsub", {
  id <- 1001
  title <- "Quotes Test"
  extent_vals <- c(0, 0, 1, 1)

  url <- zip_query_build(id, title, extent_vals = extent_vals)
  expect_false(grepl('\\"', url))  # No double quotes
  expect_true(grepl("SHP", URLdecode(url)))  # format still included
})

test_that("zip_download() returns NULL on simulated HTTP error", {
  local_mocked_bindings(
    request = function(url) list(url = url),
    req_perform = function(req) structure(list(
      body = raw(0), status_code = 404
    ), class = "response_httr2"),
    resp_status = function(resp) resp$status_code,
    resp_body_raw = function(resp) resp$body,
    .package = "httr2"
  )

  tmpf <- suppressWarnings(zip_download("http://fake-url/file.zip", id = "4d441aa75cf1"))
  expect_null(tmpf)
})

test_that("zip_download() handles network errors gracefully", {

  url <- "http://fake-url/file.zip"
  id <- "123"

  local_mocked_bindings(
    request = function(url) list(url = url),
    req_perform = function(req) stop("Network error"),
    .package = "httr2"
  )

  expect_warning(
    tmpf <- zip_download(url, id),
    regexp = "Download error for ID 123",
    fixed = FALSE
  )
  expect_null(tmpf)
})

test_that("geo_shapefiles_read() works on all extdata ZIP files", {
  skip_if_not_installed("sf")

  zip_files <- list.files(
    system.file("extdata", package = "frheritage"),
    pattern = "\\.zip$",
    full.names = FALSE
  )

  for (zip_name in zip_files) {

    zip_path <- system.file("extdata", zip_name, package = "frheritage")

    skip_if(!file.exists(zip_path))

    sf_obj <- tryCatch(
      geo_shapefiles_read(zip_path, crs = 2154),
      error = function(e) NULL
    )

    expect_true(inherits(sf_obj, "sf") || is.null(sf_obj))

    if (!is.null(sf_obj)) {
      expect_equal(sf::st_crs(sf_obj)$epsg, 2154)
    }
  }
})

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{ids_url_build}
\alias{ids_url_build}
\title{Build a search URL for the RSS service}
\usage{
ids_url_build(extent, insee_dep)
}
\arguments{
\item{extent}{A named list with elements \code{left}, \code{bottom}, \code{right}, \code{top}
defining the bounding box coordinates.}

\item{insee_dep}{A character or numeric INSEE department code.}
}
\value{
A character string containing the full URL to query the service.
}
\description{
This internal helper constructs a URL to query for the French Ministry
of Culture's "Atlas du Patrimoine" RSS service.
The URL filters data using a bounding box (\code{extent}) and an
INSEE department code (\code{insee_dep}).
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zip.R
\name{zip_query_build}
\alias{zip_query_build}
\title{Build a URL query to download a ZIP}
\usage{
zip_query_build(id, title, guid = NULL, extent_vals, crs = 2154)
}
\arguments{
\item{id}{\code{character} or \code{numeric}. Identifier of the dataset.}

\item{title}{\code{character}. Title of the dataset.}

\item{guid}{\code{character}, optional. GUID of the dataset, if available.}

\item{extent_vals}{\code{numeric} vector of length 4. Bounding box coordinates in
order: left, bottom, right, top.}

\item{crs}{\code{numeric}, default \code{2154}. CRS code for the coordinate reference
system to use in the query.}
}
\value{
\code{character}. URL string for downloading the dataset in ZIP format.
}
\description{
Constructs a URL query to request a ZIP file containing spatial data from
the French Ministry of Culture's "Atlas du Patrimoine" service.
}
\details{
\itemize{
\item The function formats the bounding box and metadata into a JSON object,
then encodes it into a URL suitable for the Atlas Patrimoines export service.
\item CRS is converted to a string \code{"EPSG:<code>"}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_heritage.R
\name{get_heritage}
\alias{get_heritage}
\title{Retrieve and download heritage spatial data for a given sf object}
\usage{
get_heritage(x, data_code, buffer = 2500, crs = 2154, verbose = TRUE)
}
\arguments{
\item{x}{An \code{sf} object defining the area of interest.}

\item{data_code}{A single \code{character} heritage dataset codes to retrieve.
Valid codes can be obtained with \code{\link[=get_heritage_layernames]{get_heritage_layernames()}}.}

\item{buffer}{A \code{numeric} value (default = 2500). Buffer distance in meters
used to slightly expand geometries before querying.}

\item{crs}{An \code{integer} or \code{sf::st_crs} object (default = 2154). Coordinate
reference system used for spatial processing.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress and diagnostic messages.}
}
\value{
A single \code{sf} object containing all retrieved heritage features.
Returns an empty \code{sf} object if no data is found or all requests fail.
}
\description{
This function retrieves and downloads spatial heritage datasets from the
French Ministry of Culture’s "Atlas du Patrimoine" service.
}
\details{
It uses the spatial extent and inferred department(s) of the input \code{sf}
object to query an internal registry of heritage datasets, then downloads
and merges the corresponding spatial layers.

This function currently supports metropolitan France departments (96 units).

Internally, the function:
\enumerate{
\item Validates the input \code{data_code} against an internal registry.
\item Prepares and optionally buffers the input geometry.
\item Computes spatial extents (bounding boxes) for each feature.
\item Assigns each feature to one or more administrative departments.
\item Filters available heritage records using an internal dataset registry.
\item Builds query URLs for the Atlas du Patrimoine service.
\item Downloads ZIP archives containing shapefiles.
\item Reads, validates, and merges the resulting \code{sf} objects.
}

Partial failures are tolerated: invalid or empty responses for individual
requests are skipped.
}
\examples{
\donttest{
if (interactive()) {

  # Define a minimal spatial object (WGS84 point)
  my_sf_layer <- sf::st_sf(
    geometry = sf::st_sfc(
      sf::st_point(c(2.21, 48.82)),
      crs = 4326
    )
  )

  # Retrieve heritage spatial data
  res <- get_heritage(
    x = my_sf_layer,
    data_code = "IMMH",
    buffer = 2000,
    crs = 2154,
    verbose = TRUE
  )

}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{extdata_examples}
\alias{extdata_examples}
\title{Example files for testing}
\description{
These files are used internally for testing functions.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{quiet}
\alias{quiet}
\title{Execute an expression completely silently}
\usage{
quiet(expr)
}
\arguments{
\item{expr}{An R expression or function call to execute silently.}
}
\value{
The result of evaluating \code{expr}.
}
\description{
This utility executes any R expression or function call and returns its value,
while suppressing all printed output (cat, print), warnings, and messages.
}
\details{
\itemize{
\item Useful to suppress noisy outputs from functions during package or script execution.
\item Evaluation occurs in the parent environment to access local variables.
\item Works cross-platform by redirecting \code{capture.output} to \code{nul} on Windows or \verb{/dev/null} on other OS.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_object_check}
\alias{geo_object_check}
\title{Check that spatial objects are valid \code{sf} layers with expected geometry type}
\usage{
geo_object_check(
  ...,
  check_class = TRUE,
  check_valid = TRUE,
  check_empty = TRUE,
  allowed_geom_classes = "POLYGON"
)
}
\arguments{
\item{...}{One or several spatial objects to test.}

\item{check_class}{Logical. Whether to check that each object inherits from
class \code{sf}. Default is \code{TRUE}.}

\item{check_valid}{Logical. Whether to check that geometries are valid.
Default is \code{TRUE}.}

\item{check_empty}{Logical. Whether to check that geometries are not empty.
Default is \code{TRUE}.}

\item{allowed_geom_classes}{Character vector of allowed geometry types
(e.g., \code{"POLYGON"}, \code{"LINE"}, \code{"POINT"}). If \code{NULL}, the geometry type
is not checked. Default is \code{"POLYGON"}.}
}
\value{
Invisibly returns \code{NULL}. Throws an error if any object fails validation.
}
\description{
This internal helper verifies that all provided objects:
\itemize{
\item are of class \code{sf} (optional, controlled by \code{check_class});
\item contain valid geometries (\code{sf::st_is_valid()});
\item contain no empty geometries (\code{!sf::st_is_empty()});
\item have a geometry type among the allowed ones (if specified).
}
}
\details{
The function raises an informative error message if any condition fails.
It is intended for internal validation of \code{sf} inputs before performing
geometric operations or API queries.

Geometry type validation is performed via \code{\link[=geo_object_type]{geo_object_type()}}, which ensures
type consistency across geometries.
}
\seealso{
\code{\link[=geo_object_type]{geo_object_type()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_heritage_ids.R
\name{get_heritage_ids}
\alias{get_heritage_ids}
\title{Retrieve heritage layer IDs for a given sf object}
\usage{
get_heritage_ids(x, buffer = 2500, crs = 2154, verbose = TRUE)
}
\arguments{
\item{x}{An \code{sf} object defining the area of interest.}

\item{buffer}{A \code{numeric} value (default = 2500). Buffer distance in meters
used to slightly expand geometries before querying.}

\item{crs}{An \code{integer} or \code{sf::st_crs} object (default = 2154). Coordinate
reference system used for spatial processing.}

\item{verbose}{Logical. If \code{TRUE} (default), prints progress and diagnostic messages.}
}
\value{
A \code{data.frame} with the following columns:
\describe{
\item{id}{Numeric identifier extracted from the record GUID.}
\item{title}{Record title as published in the service feed.}
\item{guid}{Full GUID (unique resource identifier).}
\item{departement}{Department code associated with the spatial query.}
\item{code}{Internal code associated with the layer.}
}
Returns an empty \code{data.frame} if no records are found or the request fails.
}
\description{
This function retrieves available heritage dataset identifiers from the
French Ministry of Culture’s "Atlas du Patrimoine" service.
}
\details{
It uses the spatial extent and inferred department(s) of the input \code{sf}
object to query the service.

Internally, the function:
\enumerate{
\item Aggregates nearby geometries using \code{buffer} input.
\item Determines the corresponding INSEE department code for each geometry, using \code{happign::get_wfs()}.
\item Computes the bounding box of each geometry.
\item Queries the "Atlas du Patrimoine" service feed for all available metadata
records (IDs, titles, GUIDs) within each bounding box.
}
}
\examples{
\donttest{
if (interactive()) {
  # Create a minimal sf object defining the area of interest.
  # A simple point geometry is sufficient to trigger spatial queries.
  my_sf_layer <- sf::st_sf(
    geometry = sf::st_sfc(
      sf::st_point(c(2.21, 48.82)),
      crs = 4326
    )
  )

  # Retrieve available heritage dataset identifiers intersecting
  # the spatial extent derived from the sf object.
  ids <- get_heritage_ids(my_sf_layer)
}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\docType{data}
\name{all_ids}
\alias{all_ids}
\title{Available layer datasets}
\format{
A \code{data.frame} with 5 columns:
\describe{
\item{id}{(character) Layer id.}
\item{title}{(character) Layer title.}
\item{guid}{(integer) Layer guid.}
\item{code}{(character) Internal code, determined from tile.}
\item{departement}{(character) Layer department}
}
}
\usage{
all_ids
}
\description{
A dataset containing available layer datasets from the French
Ministry of Culture's "Atlas du Patrimoine" service.
}
\details{
Contains only data available on the 96 departments of metropolitan France.
}
\keyword{datasets}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{ids_download}
\alias{ids_download}
\title{Download and parse heritage metadata from a URL}
\usage{
ids_download(url, timeout = 60)
}
\arguments{
\item{url}{\code{Character.} The complete request URL.}

\item{timeout}{\code{Numeric.} Request timeout in seconds (default = \code{60}).}
}
\value{
A \code{data.frame} with columns:
\describe{
\item{id}{Numeric identifier extracted from the GUID.}
\item{title}{Record title as published in the service.}
\item{guid}{Full GUID (unique resource identifier).}
}
Returns an empty \code{data.frame} if the request fails or no items are found.
}
\description{
This internal helper performs an HTTP GET request to the given URL
(typically generated by \code{\link[=ids_url_build]{ids_url_build()}}) and parses the RSS XML
response into a structured \code{data.frame}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_extent}
\alias{geo_extent}
\title{Compute bounding boxes for each feature in an sf object}
\usage{
geo_extent(x, crs = 4326)
}
\arguments{
\item{x}{An \code{sf} object (POLYGON or MULTIPOLYGON) for which to compute bounding boxes.}

\item{crs}{Numeric or sf CRS. Target CRS used before computation (default 4326).}
}
\value{
A \code{data.frame} with one row per feature:
\describe{
\item{left}{xmin}
\item{bottom}{ymin}
\item{right}{xmax}
\item{top}{ymax}
}
}
\description{
Computes one bounding box per geometry in an \code{sf} object.
}
\details{
\itemize{
\item The function first validates the input with \code{\link[=geo_object_check]{geo_object_check()}}.
\item Transforms the object to the target CRS before computing bounding boxes.
\item CRS and units are inherited from the input \code{sf} object.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_heritage_layernames.R
\name{get_heritage_layernames}
\alias{get_heritage_layernames}
\title{Get heritage layer codes and labels}
\usage{
get_heritage_layernames(code = NULL)
}
\arguments{
\item{code}{Optional \code{character} vector of layer codes to filter.
If \code{NULL}, the full reference table is returned.}
}
\value{
A \code{data.frame} with two columns:
\describe{
\item{code}{Official heritage layer code (character).}
\item{label}{Descriptive label (character).}
}
}
\description{
This function provides a reference table mapping internal codes to
explore heritage layer available datasets from French Ministry
of Culture's "Atlas du Patrimoine".
Optionally, it can filter the table to return only selected codes.
}
\details{
The available codes and their meanings are:
\itemize{
\item \code{IMDN} — Domaines nationaux
\item \code{IMMH} — Immeubles classes ou inscrits
\item \code{IMUN} — Patrimoine Mondial UNESCO – Emprise surfacique des biens
\item \code{LACR} — Architecture Contemporaine Remarquable
\item \code{PADN} — Protection au titre des abords de domaines nationaux
\item \code{PAMH} — Protection au titre des abords de monuments historiques
\item \code{PAUN} — Patrimoine Mondial UNESCO – Zone tampon des biens
\item \code{SICI} — Sites classes ou inscrits
\item \code{SIPR} — Sites patrimoniaux remarquables
\item \code{ZPPA} — Zones de presomption de prescriptions archeologiques
}
}
\examples{
# Get the full table of heritage layer codes
get_heritage_layernames()

# Filter for specific codes
get_heritage_layernames(c("IMMH", "SICI"))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{data_filter}
\alias{data_filter}
\title{Filter layer IDs by internal code and department}
\usage{
data_filter(ids = NULL, department, data_code)
}
\arguments{
\item{ids}{\code{data.frame}. Input dataset of IDs. \code{NULL} as default.
If \code{NULL} function uses \code{frheritage::all_ids}.}

\item{department}{\code{character} vector. Department(s) to filter.}

\item{data_code}{\code{character} vector. Data codes to filter. Must be valid heritage layer codes.}
}
\value{
\code{data.frame}. Subset of \code{ids} filtered by \code{data_code} and \code{department}, with \code{departement} column removed.
}
\description{
Filters a dataset of layer IDs based on provided data codes and departments.
}
\details{
\itemize{
\item Uses \code{data_check()} to ensure that all \code{data_code} values are valid.
\item The function currently ignores the \code{ids} argument and always uses \code{frheritage::all_ids}.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_shapefile_read}
\alias{geo_shapefile_read}
\title{Read and transform shapefile from a ZIP archive}
\usage{
geo_shapefile_read(zip_path, crs = 2154)
}
\arguments{
\item{zip_path}{\code{character}. Path to the ZIP archive to read.}

\item{crs}{\code{numeric} or \code{sf::st_crs} object. Target CRS to transform to.
Default is 2154 (Lambert-93).}
}
\value{
An \code{sf} object transformed to CRS:2154.
}
\description{
Extracts shapefiles from a ZIP archive, reads the shapefile found
in the "documents" folder into an \code{sf} object, and transforms it to
CRS:2154 (Lambert-93 projection for France), unless already in the target CRS.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_deps.R
\name{get_deps}
\alias{get_deps}
\title{Retrieve INSEE department codes from an sf object}
\usage{
get_deps(x)
}
\arguments{
\item{x}{An \code{sf} object defining the area(s) of interest.}
}
\value{
A \code{character} vector of unique INSEE department codes intersecting
the input geometries. Returns \code{NULL} if no intersection is found or if the
WFS request fails.
}
\description{
This function retrieves the INSEE administrative department codes
(\code{code_insee}) intersecting the features of a given \code{sf} object.
It relies on the \emph{Admin Express COG} WFS service provided by IGN,
accessed via the \strong{happign} package.
}
\details{
The function performs the following steps:
\enumerate{
\item Transforms the input geometries to CRS:4326.
\item Computes centroids of the geometries.
\item Queries the IGN WFS service (\emph{Admin Express COG}) to retrieve departments.
\item Performs a spatial join using \code{\link[sf:geos_binary_pred]{sf::st_intersects()}}.
\item Extracts and returns unique \code{code_insee} values.
}

Notes:
\itemize{
\item The function depends on the availability of the IGN WFS service.
\item Large or complex geometries may cause the request to fail.
\item In such cases, consider simplifying or reducing the spatial extent of \code{x}.
}
}
\examples{
\donttest{
if (interactive()) {

  # Minimal sf object (point in Paris)
  x <- sf::st_sf(
    geometry = sf::st_sfc(
      sf::st_point(c(2.35, 48.85)),
      crs = 4326
    )
  )

  # Retrieve department code
  get_deps(x)

  # Multiple locations (Paris and Rouen)
  x2 <- sf::st_sf(
    geometry = sf::st_sfc(
      sf::st_point(c(2.35, 48.85)),  # 75
      sf::st_point(c(1.09, 49.44)),  # 76
      crs = 4326
    )
  )

  get_deps(x2)

}
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{ids_to_codes}
\alias{ids_to_codes}
\title{Determine internal codes from IDs titles}
\usage{
ids_to_codes(x)
}
\arguments{
\item{x}{\code{character} vector of IDs titles.}
}
\value{
\code{character} vector of internal codes corresponding to each input IDs titles.
Returns \code{NA} if no pattern matches.
}
\description{
This internal function converts IDs tiles into their corresponding
internal codes (e.g., "IMMH", "PAMH").
It matches based on predefined keyword patterns and chooses the most
specific code when multiple patterns match.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/data.R
\name{data_check}
\alias{data_check}
\title{Check that layer data codes are valid}
\usage{
data_check(data_code)
}
\arguments{
\item{data_code}{\code{character} vector. Data codes to check.}
}
\value{
\code{NULL}. Invisibly returns \code{NULL} if all codes are valid; otherwise throws an error.
}
\description{
Validates that provided data codes exist in the heritage layers.
}
\details{
\itemize{
\item Uses \code{get_heritage_layernames()} to get the list of valid codes.
\item Throws an error if any \code{data_code} is invalid.
}
}
\seealso{
\code{\link[=get_heritage_layernames]{get_heritage_layernames()}}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_sf_bind}
\alias{geo_sf_bind}
\title{Bind multiple sf objects into a single sf}
\usage{
geo_sf_bind(sf_list)
}
\arguments{
\item{sf_list}{A list of \code{sf} objects (can include nested lists).}
}
\value{
A single \code{sf} object containing all input features, or \code{NULL}
if the input list is empty or contains no valid sf objects.
}
\description{
This internal helper takes a list of \code{sf} objects, aligns their columns,
and binds them together into a single \code{sf} object. It removes NULL or empty
objects and handles nested lists of \code{sf}.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/zip.R
\name{zip_download}
\alias{zip_download}
\title{Download a ZIP file from a given URL}
\usage{
zip_download(url, id)
}
\arguments{
\item{url}{\code{character}. URL to download the ZIP file from.}

\item{id}{\code{character} or \code{numeric}. Identifier of the dataset (used for messages).}
}
\value{
\code{character} or \code{NULL}.
Path to the temporary ZIP file, or \code{NULL} if download fails.
}
\description{
Downloads a ZIP file from a specified URL and saves it to a temporary file.
}
\details{
\itemize{
\item Uses \code{httr2} to perform the HTTP request.
\item If the HTTP response status is not 200, a warning is emitted and \code{NULL} is returned.
\item Errors during download are caught and a warning is printed.
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_cast}
\alias{geo_cast}
\title{Cast geometries to simple types}
\usage{
geo_cast(x)
}
\arguments{
\item{x}{An \code{sf} object with geometries to cast.}
}
\value{
An \code{sf} object with geometries cast to simple types.
}
\description{
This internal helper converts multi-part geometries to their single-part
equivalents: MULTIPOLYGON became POLYGON, MULTILINESTRING became LINESTRING,
MULTIPOINT became POINT. If the input contains mixed types, each feature is
cast individually.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_object_type}
\alias{geo_object_type}
\title{Determine the geometry type of an \code{sf} object}
\usage{
geo_object_type(x)
}
\arguments{
\item{x}{An \code{sf} object whose geometry type should be determined.}
}
\value{
A single \code{character(1)} value representing the normalized geometry type.
Possible values are: \code{"POINT"}, \code{"LINE"}, \code{"POLYGON"}, or \code{"GEOMETRY"}.
}
\description{
This internal helper function identifies and normalizes the geometry type
of an \code{sf} object. It ensures that all geometries within the object share
a consistent type (e.g., \code{"POINT"}, \code{"LINESTRING"}, \code{"POLYGON"}).
If multiple geometry types are found, an error is raised.
}
\details{
The function retrieves the geometry type of the input using
\code{\link[sf:st_geometry_type]{sf::st_geometry_type()}} and standardizes it according to a simplified
classification:
\itemize{
\item \code{"POINT"} and \code{"MULTIPOINT"} = \code{"POINT"}
\item \code{"LINESTRING"} and \code{"MULTILINESTRING"} = \code{"LINE"}
\item \code{"POLYGON"} and \code{"MULTIPOLYGON"} = \code{"POLYGON"}
\item All other types = \code{"GEOMETRY"}
}

Mixed or inconsistent geometry collections are not supported.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{atlas_ok}
\alias{atlas_ok}
\title{Check if the Atlas service is reachable}
\usage{
atlas_ok(url = "http://atlas.patrimoines.culture.fr/atlas/trunk/", timeout = 4)
}
\arguments{
\item{url}{A \code{character} string of length 1. The URL of the service to test.
Default is \code{"http://atlas.patrimoines.culture.fr/atlas/trunk/"}.}

\item{timeout}{Numeric. Maximum time in seconds allowed for the request
(including connection and response). Default is \code{4}.}
}
\value{
A \code{logical} scalar:
\itemize{
\item \code{TRUE} if the service responds with an HTTP status code < 400
\item \code{FALSE} otherwise (error, timeout, or HTTP status >= 400)
}
}
\description{
Performs a lightweight HTTP request to test whether a remote service
(typically the French cultural heritage Atlas) is reachable within a
specified timeout. The function uses a \code{HEAD} request to avoid downloading
content and returns a logical indicating availability.
}
\details{
This function is designed to be robust and silent: any network error
(timeout, DNS failure, connection refusal, etc.) results in \code{FALSE}.

The function relies on \pkg{httr2} and internally performs a \code{HEAD} request
using \code{httr2::req_perform()}. It does not modify global options and
does not raise errors.

Note that this function checks \strong{service reachability}, not full
functional correctness of the endpoint.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/geo.R
\name{geo_prepare}
\alias{geo_prepare}
\title{Prepare and aggregate an sf object}
\usage{
geo_prepare(x, crs = 2154, buffer = 10)
}
\arguments{
\item{x}{An \code{sf} object to process.}

\item{crs}{Integer. Code of the target CRS. Default is \code{2154}.}

\item{buffer}{Numeric. Buffer distance (in map units) used to merge nearby features. Default is \code{10}.}
}
\value{
An \code{sf} object with cleaned, projected, and aggregated geometries as \code{POLYGON}s.
}
\description{
This helper function prepares and aggregates an \code{sf} object
to produce clean, valid, and unified geometries.
}
\details{
The function performs a complete geometric preparation workflow:
\enumerate{
\item Geometry validation: ensures that all geometries belong to allowed classes (\code{POINT}, \code{LINE}, or \code{POLYGON}).
\item Geometry cleaning: invalid features are fixed with \code{\link[sf:valid]{sf::st_make_valid()}}, Z/M dimensions are dropped via \code{\link[sf:st_zm]{sf::st_zm()}}, and multipart geometries are cast to simple forms.
\item Coordinate transformation: all geometries are transformed to the specified CRS using \code{\link[sf:st_transform]{sf::st_transform()}}.
\item Aggregation: small buffer zones (controlled by \code{buffer}) are applied to merge adjacent or overlapping features.
The buffered geometries are then dissolved with \code{\link[sf:geos_combine]{sf::st_union()}} and cast to \code{POLYGON} type.
}

This process is useful for simplifying feature sets into larger contiguous study zones or analysis areas.
}
\keyword{internal}
